import json
import six

from moto.core.responses import BaseResponse

from .exceptions import SWFSerializationException, SWFValidationException
from .models import swf_backends


class SWFResponse(BaseResponse):

    @property
    def swf_backend(self):
        return swf_backends[self.region]

    # SWF parameters are passed through a JSON body, so let's ease retrieval
    @property
    def _params(self):
        return json.loads(self.body)

    def _check_int(self, parameter):
        if not isinstance(parameter, int):
            raise SWFSerializationException(parameter)

    def _check_float_or_int(self, parameter):
        if not isinstance(parameter, float):
            if not isinstance(parameter, int):
                raise SWFSerializationException(parameter)

    def _check_none_or_string(self, parameter):
        if parameter is not None:
            self._check_string(parameter)

    def _check_string(self, parameter):
        if not isinstance(parameter, six.string_types):
            raise SWFSerializationException(parameter)

    def _check_none_or_list_of_strings(self, parameter):
        if parameter is not None:
            self._check_list_of_strings(parameter)

    def _check_list_of_strings(self, parameter):
        if not isinstance(parameter, list):
            raise SWFSerializationException(parameter)
        for i in parameter:
            if not isinstance(i, six.string_types):
                raise SWFSerializationException(parameter)

    def _check_exclusivity(self, **kwargs):
        if list(kwargs.values()).count(None) >= len(kwargs) - 1:
            return
        keys = kwargs.keys()
        if len(keys) == 2:
            message = 'Cannot specify both a {0} and a {1}'.format(keys[0],
                                                                   keys[1])
        else:
            message = 'Cannot specify more than one exclusive filters in the' \
                ' same query: {0}'.format(keys)
            raise SWFValidationException(message)

    def _list_types(self, kind):
        domain_name = self._params["domain"]
        status = self._params["registrationStatus"]
        reverse_order = self._params.get("reverseOrder", None)
        self._check_string(domain_name)
        self._check_string(status)
        types = self.swf_backend.list_types(
            kind, domain_name, status, reverse_order=reverse_order)
        return json.dumps({
            "typeInfos": [_type.to_medium_dict() for _type in types]
        })

    def _describe_type(self, kind):
        domain = self._params["domain"]
        _type_args = self._params["{0}Type".format(kind)]
        name = _type_args["name"]
        version = _type_args["version"]
        self._check_string(domain)
        self._check_string(name)
        self._check_string(version)
        _type = self.swf_backend.describe_type(kind, domain, name, version)

        return json.dumps(_type.to_full_dict())

    def _deprecate_type(self, kind):
        domain = self._params["domain"]
        _type_args = self._params["{0}Type".format(kind)]
        name = _type_args["name"]
        version = _type_args["version"]
        self._check_string(domain)
        self._check_string(name)
        self._check_string(version)
        self.swf_backend.deprecate_type(kind, domain, name, version)
        return ""

    # TODO: implement pagination
    def list_domains(self):
        status = self._params["registrationStatus"]
        self._check_string(status)
        reverse_order = self._params.get("reverseOrder", None)
        domains = self.swf_backend.list_domains(
            status, reverse_order=reverse_order)
        return json.dumps({
            "domainInfos": [domain.to_short_dict() for domain in domains]
        })

    def list_closed_workflow_executions(self):
        domain = self._params['domain']
        start_time_filter = self._params.get('startTimeFilter', None)
        close_time_filter = self._params.get('closeTimeFilter', None)
        execution_filter = self._params.get('executionFilter', None)
        workflow_id = execution_filter[
            'workflowId'] if execution_filter else None
        maximum_page_size = self._params.get('maximumPageSize', 1000)
        reverse_order = self._params.get('reverseOrder', None)
        tag_filter = self._params.get('tagFilter', None)
        type_filter = self._params.get('typeFilter', None)
        close_status_filter = self._params.get('closeStatusFilter', None)

        self._check_string(domain)
        self._check_none_or_string(workflow_id)
        self._check_exclusivity(executionFilter=execution_filter,
                                typeFilter=type_filter,
                                tagFilter=tag_filter,
                                closeStatusFilter=close_status_filter)
        self._check_exclusivity(startTimeFilter=start_time_filter,
                                closeTimeFilter=close_time_filter)
        if start_time_filter is None and close_time_filter is None:
            raise SWFValidationException('Must specify time filter')
        if start_time_filter:
            self._check_float_or_int(start_time_filter['oldestDate'])
            if 'latestDate' in start_time_filter:
                self._check_float_or_int(start_time_filter['latestDate'])
        if close_time_filter:
            self._check_float_or_int(close_time_filter['oldestDate'])
            if 'latestDate' in close_time_filter:
                self._check_float_or_int(close_time_filter['latestDate'])
        if tag_filter:
            self._check_string(tag_filter['tag'])
        if type_filter:
            self._check_string(type_filter['name'])
            self._check_string(type_filter['version'])
        if close_status_filter:
            self._check_string(close_status_filter['status'])
        self._check_int(maximum_page_size)

        workflow_executions = self.swf_backend.list_closed_workflow_executions(
            domain_name=domain,
            start_time_filter=start_time_filter,
            close_time_filter=close_time_filter,
            execution_filter=execution_filter,
            tag_filter=tag_filter,
            type_filter=type_filter,
            maximum_page_size=maximum_page_size,
            reverse_order=reverse_order,
            workflow_id=workflow_id,
            close_status_filter=close_status_filter
        )

        return json.dumps({
            'executionInfos': [wfe.to_list_dict() for wfe in workflow_executions]
        })

    def list_open_workflow_executions(self):
        domain = self._params['domain']
        start_time_filter = self._params['startTimeFilter']
        execution_filter = self._params.get('executionFilter', None)
        workflow_id = execution_filter[
            'workflowId'] if execution_filter else None
        maximum_page_size = self._params.get('maximumPageSize', 1000)
        reverse_order = self._params.get('reverseOrder', None)
        tag_filter = self._params.get('tagFilter', None)
        type_filter = self._params.get('typeFilter', None)

        self._check_string(domain)
        self._check_none_or_string(workflow_id)
        self._check_exclusivity(executionFilter=execution_filter,
                                typeFilter=type_filter,
                                tagFilter=tag_filter)
        self._check_float_or_int(start_time_filter['oldestDate'])
        if 'latestDate' in start_time_filter:
            self._check_float_or_int(start_time_filter['latestDate'])
        if tag_filter:
            self._check_string(tag_filter['tag'])
        if type_filter:
            self._check_string(type_filter['name'])
            self._check_string(type_filter['version'])
        self._check_int(maximum_page_size)

        workflow_executions = self.swf_backend.list_open_workflow_executions(
            domain_name=domain,
            start_time_filter=start_time_filter,
            execution_filter=execution_filter,
            tag_filter=tag_filter,
            type_filter=type_filter,
            maximum_page_size=maximum_page_size,
            reverse_order=reverse_order,
            workflow_id=workflow_id
        )

        return json.dumps({
            'executionInfos': [wfe.to_list_dict() for wfe in workflow_executions]
        })

    def register_domain(self):
        name = self._params["name"]
        retention = self._params["workflowExecutionRetentionPeriodInDays"]
        description = self._params.get("description")
        self._check_string(retention)
        self._check_string(name)
        self._check_none_or_string(description)
        self.swf_backend.register_domain(name, retention,
                                         description=description)
        return ""

    def deprecate_domain(self):
        name = self._params["name"]
        self._check_string(name)
        self.swf_backend.deprecate_domain(name)
        return ""

    def describe_domain(self):
        name = self._params["name"]
        self._check_string(name)
        domain = self.swf_backend.describe_domain(name)
        return json.dumps(domain.to_full_dict())

    # TODO: implement pagination
    def list_activity_types(self):
        return self._list_types("activity")

    def register_activity_type(self):
        domain = self._params["domain"]
        name = self._params["name"]
        version = self._params["version"]
        default_task_list = self._params.get("defaultTaskList")
        if default_task_list:
            task_list = default_task_list.get("name")
        else:
            task_list = None
        default_task_heartbeat_timeout = self._params.get(
            "defaultTaskHeartbeatTimeout")
        default_task_schedule_to_close_timeout = self._params.get(
            "defaultTaskScheduleToCloseTimeout")
        default_task_schedule_to_start_timeout = self._params.get(
            "defaultTaskScheduleToStartTimeout")
        default_task_start_to_close_timeout = self._params.get(
            "defaultTaskStartToCloseTimeout")
        description = self._params.get("description")

        self._check_string(domain)
        self._check_string(name)
        self._check_string(version)
        self._check_none_or_string(task_list)
        self._check_none_or_string(default_task_heartbeat_timeout)
        self._check_none_or_string(default_task_schedule_to_close_timeout)
        self._check_none_or_string(default_task_schedule_to_start_timeout)
        self._check_none_or_string(default_task_start_to_close_timeout)
        self._check_none_or_string(description)

        # TODO: add defaultTaskPriority when boto gets to support it
        self.swf_backend.register_type(
            "activity", domain, name, version, task_list=task_list,
            default_task_heartbeat_timeout=default_task_heartbeat_timeout,
            default_task_schedule_to_close_timeout=default_task_schedule_to_close_timeout,
            default_task_schedule_to_start_timeout=default_task_schedule_to_start_timeout,
            default_task_start_to_close_timeout=default_task_start_to_close_timeout,
            description=description,
        )
        return ""

    def deprecate_activity_type(self):
        return self._deprecate_type("activity")

    def describe_activity_type(self):
        return self._describe_type("activity")

    def list_workflow_types(self):
        return self._list_types("workflow")

    def register_workflow_type(self):
        domain = self._params["domain"]
        name = self._params["name"]
        version = self._params["version"]
        default_task_list = self._params.get("defaultTaskList")
        if default_task_list:
            task_list = default_task_list.get("name")
        else:
            task_list = None
        default_child_policy = self._params.get("defaultChildPolicy")
        default_task_start_to_close_timeout = self._params.get(
            "defaultTaskStartToCloseTimeout")
        default_execution_start_to_close_timeout = self._params.get(
            "defaultExecutionStartToCloseTimeout")
        description = self._params.get("description")

        self._check_string(domain)
        self._check_string(name)
        self._check_string(version)
        self._check_none_or_string(task_list)
        self._check_none_or_string(default_child_policy)
        self._check_none_or_string(default_task_start_to_close_timeout)
        self._check_none_or_string(default_execution_start_to_close_timeout)
        self._check_none_or_string(description)

        # TODO: add defaultTaskPriority when boto gets to support it
        # TODO: add defaultLambdaRole when boto gets to support it
        self.swf_backend.register_type(
            "workflow", domain, name, version, task_list=task_list,
            default_child_policy=default_child_policy,
            default_task_start_to_close_timeout=default_task_start_to_close_timeout,
            default_execution_start_to_close_timeout=default_execution_start_to_close_timeout,
            description=description,
        )
        return ""

    def deprecate_workflow_type(self):
        return self._deprecate_type("workflow")

    def describe_workflow_type(self):
        return self._describe_type("workflow")

    def start_workflow_execution(self):
        domain = self._params["domain"]
        workflow_id = self._params["workflowId"]
        _workflow_type = self._params["workflowType"]
        workflow_name = _workflow_type["name"]
        workflow_version = _workflow_type["version"]
        _task_list = self._params.get("taskList")
        if _task_list:
            task_list = _task_list.get("name")
        else:
            task_list = None
        child_policy = self._params.get("childPolicy")
        execution_start_to_close_timeout = self._params.get(
            "executionStartToCloseTimeout")
        input_ = self._params.get("input")
        tag_list = self._params.get("tagList")
        task_start_to_close_timeout = self._params.get(
            "taskStartToCloseTimeout")

        self._check_string(domain)
        self._check_string(workflow_id)
        self._check_string(workflow_name)
        self._check_string(workflow_version)
        self._check_none_or_string(task_list)
        self._check_none_or_string(child_policy)
        self._check_none_or_string(execution_start_to_close_timeout)
        self._check_none_or_string(input_)
        self._check_none_or_list_of_strings(tag_list)
        self._check_none_or_string(task_start_to_close_timeout)

        wfe = self.swf_backend.start_workflow_execution(
            domain, workflow_id, workflow_name, workflow_version,
            task_list=task_list, child_policy=child_policy,
            execution_start_to_close_timeout=execution_start_to_close_timeout,
            input=input_, tag_list=tag_list,
            task_start_to_close_timeout=task_start_to_close_timeout
        )

        return json.dumps({
            "runId": wfe.run_id
        })

    def describe_workflow_execution(self):
        domain_name = self._params["domain"]
        _workflow_execution = self._params["execution"]
        run_id = _workflow_execution["runId"]
        workflow_id = _workflow_execution["workflowId"]

        self._check_string(domain_name)
        self._check_string(run_id)
        self._check_string(workflow_id)

        wfe = self.swf_backend.describe_workflow_execution(
            domain_name, run_id, workflow_id)
        return json.dumps(wfe.to_full_dict())

    def get_workflow_execution_history(self):
        domain_name = self._params["domain"]
        _workflow_execution = self._params["execution"]
        run_id = _workflow_execution["runId"]
        workflow_id = _workflow_execution["workflowId"]
        reverse_order = self._params.get("reverseOrder", None)
        wfe = self.swf_backend.describe_workflow_execution(
            domain_name, run_id, workflow_id)
        events = wfe.events(reverse_order=reverse_order)
        return json.dumps({
            "events": [evt.to_dict() for evt in events]
        })

    def poll_for_decision_task(self):
        domain_name = self._params["domain"]
        task_list = self._params["taskList"]["name"]
        identity = self._params.get("identity")
        reverse_order = self._params.get("reverseOrder", None)

        self._check_string(domain_name)
        self._check_string(task_list)

        decision = self.swf_backend.poll_for_decision_task(
            domain_name, task_list, identity=identity
        )
        if decision:
            return json.dumps(
                decision.to_full_dict(reverse_order=reverse_order)
            )
        else:
            return json.dumps({"previousStartedEventId": 0, "startedEventId": 0})

    def count_pending_decision_tasks(self):
        domain_name = self._params["domain"]
        task_list = self._params["taskList"]["name"]
        self._check_string(domain_name)
        self._check_string(task_list)
        count = self.swf_backend.count_pending_decision_tasks(
            domain_name, task_list)
        return json.dumps({"count": count, "truncated": False})

    def respond_decision_task_completed(self):
        task_token = self._params["taskToken"]
        execution_context = self._params.get("executionContext")
        decisions = self._params.get("decisions")
        self._check_string(task_token)
        self._check_none_or_string(execution_context)
        self.swf_backend.respond_decision_task_completed(
            task_token, decisions=decisions, execution_context=execution_context
        )
        return ""

    def poll_for_activity_task(self):
        domain_name = self._params["domain"]
        task_list = self._params["taskList"]["name"]
        identity = self._params.get("identity")
        self._check_string(domain_name)
        self._check_string(task_list)
        self._check_none_or_string(identity)
        activity_task = self.swf_backend.poll_for_activity_task(
            domain_name, task_list, identity=identity
        )
        if activity_task:
            return json.dumps(
                activity_task.to_full_dict()
            )
        else:
            return json.dumps({"startedEventId": 0})

    def count_pending_activity_tasks(self):
        domain_name = self._params["domain"]
        task_list = self._params["taskList"]["name"]
        self._check_string(domain_name)
        self._check_string(task_list)
        count = self.swf_backend.count_pending_activity_tasks(
            domain_name, task_list)
        return json.dumps({"count": count, "truncated": False})

    def respond_activity_task_completed(self):
        task_token = self._params["taskToken"]
        result = self._params.get("result")
        self._check_string(task_token)
        self._check_none_or_string(result)
        self.swf_backend.respond_activity_task_completed(
            task_token, result=result
        )
        return ""

    def respond_activity_task_failed(self):
        task_token = self._params["taskToken"]
        reason = self._params.get("reason")
        details = self._params.get("details")
        self._check_string(task_token)
        # TODO: implement length limits on reason and details (common pb with
        # client libs)
        self._check_none_or_string(reason)
        self._check_none_or_string(details)
        self.swf_backend.respond_activity_task_failed(
            task_token, reason=reason, details=details
        )
        return ""

    def terminate_workflow_execution(self):
        domain_name = self._params["domain"]
        workflow_id = self._params["workflowId"]
        child_policy = self._params.get("childPolicy")
        details = self._params.get("details")
        reason = self._params.get("reason")
        run_id = self._params.get("runId")
        self._check_string(domain_name)
        self._check_string(workflow_id)
        self._check_none_or_string(child_policy)
        self._check_none_or_string(details)
        self._check_none_or_string(reason)
        self._check_none_or_string(run_id)
        self.swf_backend.terminate_workflow_execution(
            domain_name, workflow_id, child_policy=child_policy,
            details=details, reason=reason, run_id=run_id
        )
        return ""

    def record_activity_task_heartbeat(self):
        task_token = self._params["taskToken"]
        details = self._params.get("details")
        self._check_string(task_token)
        self._check_none_or_string(details)
        self.swf_backend.record_activity_task_heartbeat(
            task_token, details=details
        )
        # TODO: make it dynamic when we implement activity tasks cancellation
        return json.dumps({"cancelRequested": False})

    def signal_workflow_execution(self):
        domain_name = self._params["domain"]
        signal_name = self._params["signalName"]
        workflow_id = self._params["workflowId"]
        _input = self._params["input"]
        run_id = self._params["runId"]

        self._check_string(domain_name)
        self._check_string(signal_name)
        self._check_string(workflow_id)
        self._check_none_or_string(_input)
        self._check_none_or_string(run_id)

        self.swf_backend.signal_workflow_execution(
            domain_name, signal_name, workflow_id, _input, run_id)
        return ""



from __future__ import unicode_literals

import sure  # noqa

import moto.server as server
from moto import mock_polly

'''
Test the different server responses
'''


@mock_polly
def test_polly_list():
    backend = server.create_backend_app("polly")
    test_client = backend.test_client()

    res = test_client.get('/v1/lexicons')
    res.status_code.should.equal(200)



from __future__ import unicode_literals

import boto
import boto3
from boto3.dynamodb.conditions import Key
import sure  # noqa
from freezegun import freeze_time
from boto.exception import JSONResponseError
from moto import mock_dynamodb2, mock_dynamodb2_deprecated
from tests.helpers import requires_boto_gte
import botocore
try:
    from boto.dynamodb2.fields import HashKey
    from boto.dynamodb2.table import Table
    from boto.dynamodb2.table import Item
    from boto.dynamodb2.exceptions import ConditionalCheckFailedException, ItemNotFound
except ImportError:
    pass


def create_table():
    table = Table.create('messages', schema=[
        HashKey('forum_name')
    ], throughput={
        'read': 10,
        'write': 10,
    })
    return table


@requires_boto_gte("2.9")
@mock_dynamodb2_deprecated
@freeze_time("2012-01-14")
def test_create_table():
    create_table()
    expected = {
        'Table': {
            'AttributeDefinitions': [
                {'AttributeName': 'forum_name', 'AttributeType': 'S'}
            ],
            'ProvisionedThroughput': {
                'NumberOfDecreasesToday': 0, 'WriteCapacityUnits': 10, 'ReadCapacityUnits': 10
            },
            'TableSizeBytes': 0,
            'TableName': 'messages',
            'TableStatus': 'ACTIVE',
            'TableArn': 'arn:aws:dynamodb:us-east-1:123456789011:table/messages',
            'KeySchema': [
                {'KeyType': 'HASH', 'AttributeName': 'forum_name'}
            ],
            'ItemCount': 0, 'CreationDateTime': 1326499200.0,
            'GlobalSecondaryIndexes': [],
            'LocalSecondaryIndexes': []
        }
    }
    conn = boto.dynamodb2.connect_to_region(
        'us-east-1',
        aws_access_key_id="ak",
        aws_secret_access_key="sk"
    )

    conn.describe_table('messages').should.equal(expected)


@requires_boto_gte("2.9")
@mock_dynamodb2_deprecated
def test_delete_table():
    create_table()
    conn = boto.dynamodb2.layer1.DynamoDBConnection()
    conn.list_tables()["TableNames"].should.have.length_of(1)

    conn.delete_table('messages')
    conn.list_tables()["TableNames"].should.have.length_of(0)

    conn.delete_table.when.called_with(
        'messages').should.throw(JSONResponseError)


@requires_boto_gte("2.9")
@mock_dynamodb2_deprecated
def test_update_table_throughput():
    table = create_table()
    table.throughput["read"].should.equal(10)
    table.throughput["write"].should.equal(10)

    table.update(throughput={
        'read': 5,
        'write': 6,
    })

    table.throughput["read"].should.equal(5)
    table.throughput["write"].should.equal(6)


@requires_boto_gte("2.9")
@mock_dynamodb2_deprecated
def test_item_add_and_describe_and_update():
    table = create_table()

    data = {
        'forum_name': 'LOLCat Forum',
        'Body': 'http://url_to_lolcat.gif',
        'SentBy': 'User A',
    }

    table.put_item(data=data)
    returned_item = table.get_item(forum_name="LOLCat Forum")
    returned_item.should_not.be.none

    dict(returned_item).should.equal({
        'forum_name': 'LOLCat Forum',
        'Body': 'http://url_to_lolcat.gif',
        'SentBy': 'User A',
    })

    returned_item['SentBy'] = 'User B'
    returned_item.save(overwrite=True)

    returned_item = table.get_item(
        forum_name='LOLCat Forum'
    )
    dict(returned_item).should.equal({
        'forum_name': 'LOLCat Forum',
        'Body': 'http://url_to_lolcat.gif',
        'SentBy': 'User B',
    })


@requires_boto_gte("2.9")
@mock_dynamodb2_deprecated
def test_item_partial_save():
    table = create_table()

    data = {
        'forum_name': 'LOLCat Forum',
        'Body': 'http://url_to_lolcat.gif',
        'SentBy': 'User A',
    }

    table.put_item(data=data)
    returned_item = table.get_item(forum_name="LOLCat Forum")

    returned_item['SentBy'] = 'User B'
    returned_item.partial_save()

    returned_item = table.get_item(
        forum_name='LOLCat Forum'
    )
    dict(returned_item).should.equal({
        'forum_name': 'LOLCat Forum',
        'Body': 'http://url_to_lolcat.gif',
        'SentBy': 'User B',
    })


@requires_boto_gte("2.9")
@mock_dynamodb2_deprecated
def test_item_put_without_table():
    conn = boto.dynamodb2.layer1.DynamoDBConnection()

    conn.put_item.when.called_with(
        table_name='undeclared-table',
        item={
            'forum_name': 'LOLCat Forum',
            'Body': 'http://url_to_lolcat.gif',
            'SentBy': 'User A',
        }
    ).should.throw(JSONResponseError)


@requires_boto_gte("2.9")
@mock_dynamodb2_deprecated
def test_get_item_with_undeclared_table():
    conn = boto.dynamodb2.layer1.DynamoDBConnection()

    conn.get_item.when.called_with(
        table_name='undeclared-table',
        key={"forum_name": {"S": "LOLCat Forum"}},
    ).should.throw(JSONResponseError)


@requires_boto_gte("2.30.0")
@mock_dynamodb2_deprecated
def test_delete_item():
    table = create_table()

    item_data = {
        'forum_name': 'LOLCat Forum',
        'Body': 'http://url_to_lolcat.gif',
        'SentBy': 'User A',
        'ReceivedTime': '12/9/2011 11:36:03 PM',
    }
    item = Item(table, item_data)
    item.save()
    table.count().should.equal(1)

    response = item.delete()

    response.should.equal(True)

    table.count().should.equal(0)

    # Deletes are idempotent and 'False' here would imply an error condition
    item.delete().should.equal(True)


@requires_boto_gte("2.9")
@mock_dynamodb2_deprecated
def test_delete_item_with_undeclared_table():
    conn = boto.dynamodb2.layer1.DynamoDBConnection()

    conn.delete_item.when.called_with(
        table_name='undeclared-table',
        key={"forum_name": {"S": "LOLCat Forum"}},
    ).should.throw(JSONResponseError)


@requires_boto_gte("2.9")
@mock_dynamodb2_deprecated
def test_query():
    table = create_table()

    item_data = {
        'forum_name': 'the-key',
        'Body': 'http://url_to_lolcat.gif',
        'SentBy': 'User A',
        'ReceivedTime': '12/9/2011 11:36:03 PM',
    }
    item = Item(table, item_data)
    item.save(overwrite=True)
    table.count().should.equal(1)
    table = Table("messages")

    results = table.query(forum_name__eq='the-key')
    sum(1 for _ in results).should.equal(1)


@requires_boto_gte("2.9")
@mock_dynamodb2_deprecated
def test_query_with_undeclared_table():
    conn = boto.dynamodb2.layer1.DynamoDBConnection()

    conn.query.when.called_with(
        table_name='undeclared-table',
        key_conditions={"forum_name": {
            "ComparisonOperator": "EQ", "AttributeValueList": [{"S": "the-key"}]}}
    ).should.throw(JSONResponseError)


@requires_boto_gte("2.9")
@mock_dynamodb2_deprecated
def test_scan():
    table = create_table()

    item_data = {
        'Body': 'http://url_to_lolcat.gif',
        'SentBy': 'User A',
        'ReceivedTime': '12/9/2011 11:36:03 PM',
    }
    item_data['forum_name'] = 'the-key'

    item = Item(table, item_data)
    item.save()

    item['forum_name'] = 'the-key2'
    item.save(overwrite=True)

    item_data = {
        'Body': 'http://url_to_lolcat.gif',
        'SentBy': 'User B',
        'ReceivedTime': '12/9/2011 11:36:03 PM',
        'Ids': set([1, 2, 3]),
        'PK': 7,
    }
    item_data['forum_name'] = 'the-key3'
    item = Item(table, item_data)
    item.save()

    results = table.scan()
    sum(1 for _ in results).should.equal(3)

    results = table.scan(SentBy__eq='User B')
    sum(1 for _ in results).should.equal(1)

    results = table.scan(Body__beginswith='http')
    sum(1 for _ in results).should.equal(3)

    results = table.scan(Ids__null=False)
    sum(1 for _ in results).should.equal(1)

    results = table.scan(Ids__null=True)
    sum(1 for _ in results).should.equal(2)

    results = table.scan(PK__between=[8, 9])
    sum(1 for _ in results).should.equal(0)

    results = table.scan(PK__between=[5, 8])
    sum(1 for _ in results).should.equal(1)


@requires_boto_gte("2.9")
@mock_dynamodb2_deprecated
def test_scan_with_undeclared_table():
    conn = boto.dynamodb2.layer1.DynamoDBConnection()

    conn.scan.when.called_with(
        table_name='undeclared-table',
        scan_filter={
            "SentBy": {
                "AttributeValueList": [{
                    "S": "User B"}
                ],
                "ComparisonOperator": "EQ"
            }
        },
    ).should.throw(JSONResponseError)


@requires_boto_gte("2.9")
@mock_dynamodb2_deprecated
def test_write_batch():
    table = create_table()

    with table.batch_write() as batch:
        batch.put_item(data={
            'forum_name': 'the-key',
            'subject': '123',
            'Body': 'http://url_to_lolcat.gif',
            'SentBy': 'User A',
            'ReceivedTime': '12/9/2011 11:36:03 PM',
        })
        batch.put_item(data={
            'forum_name': 'the-key2',
            'subject': '789',
            'Body': 'http://url_to_lolcat.gif',
            'SentBy': 'User B',
            'ReceivedTime': '12/9/2011 11:36:03 PM',
        })

    table.count().should.equal(2)
    with table.batch_write() as batch:
        batch.delete_item(
            forum_name='the-key',
            subject='789'
        )

    table.count().should.equal(1)


@requires_boto_gte("2.9")
@mock_dynamodb2_deprecated
def test_batch_read():
    table = create_table()

    item_data = {
        'Body': 'http://url_to_lolcat.gif',
        'SentBy': 'User A',
        'ReceivedTime': '12/9/2011 11:36:03 PM',
    }
    item_data['forum_name'] = 'the-key1'
    item = Item(table, item_data)
    item.save()

    item = Item(table, item_data)
    item_data['forum_name'] = 'the-key2'
    item.save(overwrite=True)

    item_data = {
        'Body': 'http://url_to_lolcat.gif',
        'SentBy': 'User B',
        'ReceivedTime': '12/9/2011 11:36:03 PM',
        'Ids': set([1, 2, 3]),
        'PK': 7,
    }
    item = Item(table, item_data)
    item_data['forum_name'] = 'another-key'
    item.save(overwrite=True)

    results = table.batch_get(
        keys=[
            {'forum_name': 'the-key1'},
            {'forum_name': 'another-key'},
        ]
    )

    # Iterate through so that batch_item gets called
    count = len([x for x in results])
    count.should.equal(2)


@requires_boto_gte("2.9")
@mock_dynamodb2_deprecated
def test_get_key_fields():
    table = create_table()
    kf = table.get_key_fields()
    kf[0].should.equal('forum_name')


@requires_boto_gte("2.9")
@mock_dynamodb2_deprecated
def test_get_missing_item():
    table = create_table()
    table.get_item.when.called_with(
        forum_name='missing').should.throw(ItemNotFound)


@requires_boto_gte("2.9")
@mock_dynamodb2_deprecated
def test_get_special_item():
    table = Table.create('messages', schema=[
        HashKey('date-joined')
    ], throughput={
        'read': 10,
        'write': 10,
    })

    data = {
        'date-joined': 127549192,
        'SentBy': 'User A',
    }
    table.put_item(data=data)
    returned_item = table.get_item(**{'date-joined': 127549192})
    dict(returned_item).should.equal(data)


@mock_dynamodb2_deprecated
def test_update_item_remove():
    conn = boto.dynamodb2.connect_to_region("us-east-1")
    table = Table.create('messages', schema=[
        HashKey('username')
    ])

    data = {
        'username': "steve",
        'SentBy': 'User A',
        'SentTo': 'User B',
    }
    table.put_item(data=data)
    key_map = {
        'username': {"S": "steve"}
    }

    # Then remove the SentBy field
    conn.update_item("messages", key_map,
                     update_expression="REMOVE SentBy, SentTo")

    returned_item = table.get_item(username="steve")
    dict(returned_item).should.equal({
        'username': "steve",
    })


@mock_dynamodb2_deprecated
def test_update_item_set():
    conn = boto.dynamodb2.connect_to_region("us-east-1")
    table = Table.create('messages', schema=[
        HashKey('username')
    ])

    data = {
        'username': "steve",
        'SentBy': 'User A',
    }
    table.put_item(data=data)
    key_map = {
        'username': {"S": "steve"}
    }

    conn.update_item("messages", key_map,
                     update_expression="SET foo=bar, blah=baz REMOVE SentBy")

    returned_item = table.get_item(username="steve")
    dict(returned_item).should.equal({
        'username': "steve",
        'foo': 'bar',
        'blah': 'baz',
    })


@mock_dynamodb2_deprecated
def test_failed_overwrite():
    table = Table.create('messages', schema=[
        HashKey('id'),
    ], throughput={
        'read': 7,
        'write': 3,
    })

    data1 = {'id': '123', 'data': '678'}
    table.put_item(data=data1)

    data2 = {'id': '123', 'data': '345'}
    table.put_item(data=data2, overwrite=True)

    data3 = {'id': '123', 'data': '812'}
    table.put_item.when.called_with(data=data3).should.throw(
        ConditionalCheckFailedException)

    returned_item = table.lookup('123')
    dict(returned_item).should.equal(data2)

    data4 = {'id': '124', 'data': 812}
    table.put_item(data=data4)

    returned_item = table.lookup('124')
    dict(returned_item).should.equal(data4)


@mock_dynamodb2_deprecated
def test_conflicting_writes():
    table = Table.create('messages', schema=[
        HashKey('id'),
    ])

    item_data = {'id': '123', 'data': '678'}
    item1 = Item(table, item_data)
    item2 = Item(table, item_data)
    item1.save()

    item1['data'] = '579'
    item2['data'] = '912'

    item1.save()
    item2.save.when.called_with().should.throw(ConditionalCheckFailedException)


"""
boto3
"""


@mock_dynamodb2
def test_boto3_create_table():
    dynamodb = boto3.resource('dynamodb', region_name='us-east-1')

    table = dynamodb.create_table(
        TableName='users',
        KeySchema=[
            {
                'AttributeName': 'username',
                'KeyType': 'HASH'
            },
        ],
        AttributeDefinitions=[
            {
                'AttributeName': 'username',
                'AttributeType': 'S'
            },
        ],
        ProvisionedThroughput={
            'ReadCapacityUnits': 5,
            'WriteCapacityUnits': 5
        }
    )
    table.name.should.equal('users')


def _create_user_table():
    dynamodb = boto3.resource('dynamodb', region_name='us-east-1')

    table = dynamodb.create_table(
        TableName='users',
        KeySchema=[
            {
                'AttributeName': 'username',
                'KeyType': 'HASH'
            },
        ],
        AttributeDefinitions=[
            {
                'AttributeName': 'username',
                'AttributeType': 'S'
            },
        ],
        ProvisionedThroughput={
            'ReadCapacityUnits': 5,
            'WriteCapacityUnits': 5
        }
    )
    return dynamodb.Table('users')


@mock_dynamodb2
def test_boto3_conditions():
    table = _create_user_table()

    table.put_item(Item={'username': 'johndoe'})
    table.put_item(Item={'username': 'janedoe'})

    response = table.query(
        KeyConditionExpression=Key('username').eq('johndoe')
    )
    response['Count'].should.equal(1)
    response['Items'].should.have.length_of(1)
    response['Items'][0].should.equal({"username": "johndoe"})


@mock_dynamodb2
def test_boto3_put_item_conditions_pass():
    table = _create_user_table()
    table.put_item(Item={'username': 'johndoe', 'foo': 'bar'})
    table.put_item(
        Item={'username': 'johndoe', 'foo': 'baz'},
        Expected={
            'foo': {
                'ComparisonOperator': 'EQ',
                'AttributeValueList': ['bar']
            }
        })
    final_item = table.get_item(Key={'username': 'johndoe'})
    assert dict(final_item)['Item']['foo'].should.equal("baz")

@mock_dynamodb2
def test_boto3_put_item_conditions_pass_because_expect_not_exists_by_compare_to_null():
    table = _create_user_table()
    table.put_item(Item={'username': 'johndoe', 'foo': 'bar'})
    table.put_item(
        Item={'username': 'johndoe', 'foo': 'baz'},
        Expected={
            'whatever': {
                'ComparisonOperator': 'NULL',
            }
        })
    final_item = table.get_item(Key={'username': 'johndoe'})
    assert dict(final_item)['Item']['foo'].should.equal("baz")

@mock_dynamodb2
def test_boto3_put_item_conditions_pass_because_expect_exists_by_compare_to_not_null():
    table = _create_user_table()
    table.put_item(Item={'username': 'johndoe', 'foo': 'bar'})
    table.put_item(
        Item={'username': 'johndoe', 'foo': 'baz'},
        Expected={
            'foo': {
                'ComparisonOperator': 'NOT_NULL',
            }
        })
    final_item = table.get_item(Key={'username': 'johndoe'})
    assert dict(final_item)['Item']['foo'].should.equal("baz")

@mock_dynamodb2
def test_boto3_put_item_conditions_fail():
    table = _create_user_table()
    table.put_item(Item={'username': 'johndoe', 'foo': 'bar'})
    table.put_item.when.called_with(
        Item={'username': 'johndoe', 'foo': 'baz'},
        Expected={
            'foo': {
                'ComparisonOperator': 'NE',
                'AttributeValueList': ['bar']
            }
        }).should.throw(botocore.client.ClientError)

@mock_dynamodb2
def test_boto3_update_item_conditions_fail():
    table = _create_user_table()
    table.put_item(Item={'username': 'johndoe', 'foo': 'baz'})
    table.update_item.when.called_with(
        Key={'username': 'johndoe'},
        UpdateExpression='SET foo=bar',
        Expected={
            'foo': {
                'Value': 'bar',
            }
        }).should.throw(botocore.client.ClientError)

@mock_dynamodb2
def test_boto3_update_item_conditions_fail_because_expect_not_exists():
    table = _create_user_table()
    table.put_item(Item={'username': 'johndoe', 'foo': 'baz'})
    table.update_item.when.called_with(
        Key={'username': 'johndoe'},
        UpdateExpression='SET foo=bar',
        Expected={
            'foo': {
                'Exists': False
            }
        }).should.throw(botocore.client.ClientError)

@mock_dynamodb2
def test_boto3_update_item_conditions_fail_because_expect_not_exists_by_compare_to_null():
    table = _create_user_table()
    table.put_item(Item={'username': 'johndoe', 'foo': 'baz'})
    table.update_item.when.called_with(
        Key={'username': 'johndoe'},
        UpdateExpression='SET foo=bar',
        Expected={
            'foo': {
                'ComparisonOperator': 'NULL',
            }
        }).should.throw(botocore.client.ClientError)

@mock_dynamodb2
def test_boto3_update_item_conditions_pass():
    table = _create_user_table()
    table.put_item(Item={'username': 'johndoe', 'foo': 'bar'})
    table.update_item(
        Key={'username': 'johndoe'},
        UpdateExpression='SET foo=baz',
        Expected={
            'foo': {
                'Value': 'bar',
            }
        })
    returned_item = table.get_item(Key={'username': 'johndoe'})
    assert dict(returned_item)['Item']['foo'].should.equal("baz")

@mock_dynamodb2
def test_boto3_update_item_conditions_pass_because_expect_not_exists():
    table = _create_user_table()
    table.put_item(Item={'username': 'johndoe', 'foo': 'bar'})
    table.update_item(
        Key={'username': 'johndoe'},
        UpdateExpression='SET foo=baz',
        Expected={
            'whatever': {
                'Exists': False,
            }
        })
    returned_item = table.get_item(Key={'username': 'johndoe'})
    assert dict(returned_item)['Item']['foo'].should.equal("baz")

@mock_dynamodb2
def test_boto3_update_item_conditions_pass_because_expect_not_exists_by_compare_to_null():
    table = _create_user_table()
    table.put_item(Item={'username': 'johndoe', 'foo': 'bar'})
    table.update_item(
        Key={'username': 'johndoe'},
        UpdateExpression='SET foo=baz',
        Expected={
            'whatever': {
                'ComparisonOperator': 'NULL',
            }
        })
    returned_item = table.get_item(Key={'username': 'johndoe'})
    assert dict(returned_item)['Item']['foo'].should.equal("baz")

@mock_dynamodb2
def test_boto3_update_item_conditions_pass_because_expect_exists_by_compare_to_not_null():
    table = _create_user_table()
    table.put_item(Item={'username': 'johndoe', 'foo': 'bar'})
    table.update_item(
        Key={'username': 'johndoe'},
        UpdateExpression='SET foo=baz',
        Expected={
            'foo': {
                'ComparisonOperator': 'NOT_NULL',
            }
        })
    returned_item = table.get_item(Key={'username': 'johndoe'})
    assert dict(returned_item)['Item']['foo'].should.equal("baz")


@mock_dynamodb2
def test_boto3_update_settype_item_with_conditions():
    class OrderedSet(set):
        """A set with predictable iteration order"""
        def __init__(self, values):
            super(OrderedSet, self).__init__(values)
            self.__ordered_values = values

        def __iter__(self):
            return iter(self.__ordered_values)

    table = _create_user_table()
    table.put_item(Item={'username': 'johndoe'})
    table.update_item(
        Key={'username': 'johndoe'},
        UpdateExpression='SET foo=:new_value',
        ExpressionAttributeValues={
            ':new_value': OrderedSet(['hello', 'world']),
        },
    )

    table.update_item(
        Key={'username': 'johndoe'},
        UpdateExpression='SET foo=:new_value',
        ExpressionAttributeValues={
            ':new_value': set(['baz']),
        },
        Expected={
            'foo': {
                'ComparisonOperator': 'EQ',
                'AttributeValueList': [
                    OrderedSet(['world', 'hello']),  # Opposite order to original
                ],
            }
        },
    )
    returned_item = table.get_item(Key={'username': 'johndoe'})
    assert dict(returned_item)['Item']['foo'].should.equal(set(['baz']))


@mock_dynamodb2
def test_boto3_put_item_conditions_pass():
    table = _create_user_table()
    table.put_item(Item={'username': 'johndoe', 'foo': 'bar'})
    table.put_item(
        Item={'username': 'johndoe', 'foo': 'baz'},
        Expected={
            'foo': {
                'ComparisonOperator': 'EQ',
                'AttributeValueList': ['bar']
            }
        })
    returned_item = table.get_item(Key={'username': 'johndoe'})
    assert dict(returned_item)['Item']['foo'].should.equal("baz")


@mock_dynamodb2
def test_scan_pagination():
    table = _create_user_table()

    expected_usernames = ['user{0}'.format(i) for i in range(10)]
    for u in expected_usernames:
        table.put_item(Item={'username': u})

    page1 = table.scan(Limit=6)
    page1['Count'].should.equal(6)
    page1['Items'].should.have.length_of(6)
    page1.should.have.key('LastEvaluatedKey')

    page2 = table.scan(Limit=6,
                       ExclusiveStartKey=page1['LastEvaluatedKey'])
    page2['Count'].should.equal(4)
    page2['Items'].should.have.length_of(4)
    page2.should_not.have.key('LastEvaluatedKey')

    results = page1['Items'] + page2['Items']
    usernames = set([r['username'] for r in results])
    usernames.should.equal(set(expected_usernames))


@mock_dynamodb2
def test_scan_by_index():
    dynamodb = boto3.client('dynamodb', region_name='us-east-1')

    dynamodb.create_table(
        TableName='test',
        KeySchema=[{'AttributeName': 'id', 'KeyType': 'HASH'}],
        AttributeDefinitions=[
            {'AttributeName': 'id', 'AttributeType': 'S'},
            {'AttributeName': 'gsi_col', 'AttributeType': 'S'}
        ],
        ProvisionedThroughput={'ReadCapacityUnits': 1, 'WriteCapacityUnits': 1},
        GlobalSecondaryIndexes=[
            {
                'IndexName': 'test_gsi',
                'KeySchema': [
                    {
                        'AttributeName': 'gsi_col',
                        'KeyType': 'HASH'
                    },
                ],
                'Projection': {
                    'ProjectionType': 'ALL',
                },
                'ProvisionedThroughput': {
                    'ReadCapacityUnits': 1,
                    'WriteCapacityUnits': 1
                }
            },
        ]
    )

    dynamodb.put_item(
        TableName='test',
        Item={
            'id': {'S': '1'},
            'col1': {'S': 'val1'},
            'gsi_col': {'S': 'gsi_val1'},
        }
    )

    dynamodb.put_item(
        TableName='test',
        Item={
            'id': {'S': '2'},
            'col1': {'S': 'val2'},
            'gsi_col': {'S': 'gsi_val2'},
        }
    )

    dynamodb.put_item(
        TableName='test',
        Item={
            'id': {'S': '3'},
            'col1': {'S': 'val3'},
        }
    )

    res = dynamodb.scan(TableName='test')
    assert res['Count'] == 3
    assert len(res['Items']) == 3

    res = dynamodb.scan(TableName='test', IndexName='test_gsi')
    assert res['Count'] == 2
    assert len(res['Items']) == 2

    res = dynamodb.scan(TableName='test', IndexName='test_gsi', Limit=1)
    assert res['Count'] == 1
    assert len(res['Items']) == 1
    last_eval_key = res['LastEvaluatedKey']
    assert last_eval_key['id']['S'] == '1'
    assert last_eval_key['gsi_col']['S'] == 'gsi_val1'



from __future__ import unicode_literals

import os
import boto3
from freezegun import freeze_time
import sure  # noqa
import uuid

from botocore.exceptions import ClientError

from moto import mock_acm


RESOURCE_FOLDER = os.path.join(os.path.dirname(__file__), 'resources')
_GET_RESOURCE = lambda x: open(os.path.join(RESOURCE_FOLDER, x), 'rb').read()
CA_CRT = _GET_RESOURCE('ca.pem')
CA_KEY = _GET_RESOURCE('ca.key')
SERVER_CRT = _GET_RESOURCE('star_moto_com.pem')
SERVER_COMMON_NAME = '*.moto.com'
SERVER_CRT_BAD = _GET_RESOURCE('star_moto_com-bad.pem')
SERVER_KEY = _GET_RESOURCE('star_moto_com.key')
BAD_ARN = 'arn:aws:acm:us-east-2:123456789012:certificate/_0000000-0000-0000-0000-000000000000'


def _import_cert(client):
    response = client.import_certificate(
        Certificate=SERVER_CRT,
        PrivateKey=SERVER_KEY,
        CertificateChain=CA_CRT
    )
    return response['CertificateArn']


# Also tests GetCertificate
@mock_acm
def test_import_certificate():
    client = boto3.client('acm', region_name='eu-central-1')

    resp = client.import_certificate(
        Certificate=SERVER_CRT,
        PrivateKey=SERVER_KEY,
        CertificateChain=CA_CRT
    )
    resp = client.get_certificate(CertificateArn=resp['CertificateArn'])

    resp['Certificate'].should.equal(SERVER_CRT.decode())
    resp.should.contain('CertificateChain')


@mock_acm
def test_import_bad_certificate():
    client = boto3.client('acm', region_name='eu-central-1')

    try:
        client.import_certificate(
            Certificate=SERVER_CRT_BAD,
            PrivateKey=SERVER_KEY,
        )
    except ClientError as err:
        err.response['Error']['Code'].should.equal('ValidationException')
    else:
        raise RuntimeError('Should of raised ValidationException')


@mock_acm
def test_list_certificates():
    client = boto3.client('acm', region_name='eu-central-1')
    arn = _import_cert(client)

    resp = client.list_certificates()
    len(resp['CertificateSummaryList']).should.equal(1)

    resp['CertificateSummaryList'][0]['CertificateArn'].should.equal(arn)
    resp['CertificateSummaryList'][0]['DomainName'].should.equal(SERVER_COMMON_NAME)


@mock_acm
def test_get_invalid_certificate():
    client = boto3.client('acm', region_name='eu-central-1')

    try:
        client.get_certificate(CertificateArn=BAD_ARN)
    except ClientError as err:
        err.response['Error']['Code'].should.equal('ResourceNotFoundException')
    else:
        raise RuntimeError('Should of raised ResourceNotFoundException')


# Also tests deleting invalid certificate
@mock_acm
def test_delete_certificate():
    client = boto3.client('acm', region_name='eu-central-1')
    arn = _import_cert(client)

    # If it does not raise an error and the next call does, all is fine
    client.delete_certificate(CertificateArn=arn)

    try:
        client.delete_certificate(CertificateArn=arn)
    except ClientError as err:
        err.response['Error']['Code'].should.equal('ResourceNotFoundException')
    else:
        raise RuntimeError('Should of raised ResourceNotFoundException')


@mock_acm
def test_describe_certificate():
    client = boto3.client('acm', region_name='eu-central-1')
    arn = _import_cert(client)

    resp = client.describe_certificate(CertificateArn=arn)
    resp['Certificate']['CertificateArn'].should.equal(arn)
    resp['Certificate']['DomainName'].should.equal(SERVER_COMMON_NAME)
    resp['Certificate']['Issuer'].should.equal('Moto')
    resp['Certificate']['KeyAlgorithm'].should.equal('RSA_2048')
    resp['Certificate']['Status'].should.equal('ISSUED')
    resp['Certificate']['Type'].should.equal('IMPORTED')


@mock_acm
def test_describe_certificate():
    client = boto3.client('acm', region_name='eu-central-1')

    try:
        client.describe_certificate(CertificateArn=BAD_ARN)
    except ClientError as err:
        err.response['Error']['Code'].should.equal('ResourceNotFoundException')
    else:
        raise RuntimeError('Should of raised ResourceNotFoundException')


# Also tests ListTagsForCertificate
@mock_acm
def test_add_tags_to_certificate():
    client = boto3.client('acm', region_name='eu-central-1')
    arn = _import_cert(client)

    client.add_tags_to_certificate(
        CertificateArn=arn,
        Tags=[
            {'Key': 'key1', 'Value': 'value1'},
            {'Key': 'key2'},
        ]
    )

    resp = client.list_tags_for_certificate(CertificateArn=arn)
    tags = {item['Key']: item.get('Value', '__NONE__') for item in resp['Tags']}

    tags.should.contain('key1')
    tags.should.contain('key2')
    tags['key1'].should.equal('value1')

    # This way, it ensures that we can detect if None is passed back when it shouldnt,
    # as we store keys without values with a value of None, but it shouldnt be passed back
    tags['key2'].should.equal('__NONE__')


@mock_acm
def test_add_tags_to_invalid_certificate():
    client = boto3.client('acm', region_name='eu-central-1')

    try:
        client.add_tags_to_certificate(
            CertificateArn=BAD_ARN,
            Tags=[
                {'Key': 'key1', 'Value': 'value1'},
                {'Key': 'key2'},
            ]
        )
    except ClientError as err:
        err.response['Error']['Code'].should.equal('ResourceNotFoundException')
    else:
        raise RuntimeError('Should of raised ResourceNotFoundException')


@mock_acm
def test_list_tags_for_invalid_certificate():
    client = boto3.client('acm', region_name='eu-central-1')

    try:
        client.list_tags_for_certificate(CertificateArn=BAD_ARN)
    except ClientError as err:
        err.response['Error']['Code'].should.equal('ResourceNotFoundException')
    else:
        raise RuntimeError('Should of raised ResourceNotFoundException')


@mock_acm
def test_remove_tags_from_certificate():
    client = boto3.client('acm', region_name='eu-central-1')
    arn = _import_cert(client)

    client.add_tags_to_certificate(
        CertificateArn=arn,
        Tags=[
            {'Key': 'key1', 'Value': 'value1'},
            {'Key': 'key2'},
            {'Key': 'key3', 'Value': 'value3'},
            {'Key': 'key4', 'Value': 'value4'},
        ]
    )

    client.remove_tags_from_certificate(
        CertificateArn=arn,
        Tags=[
            {'Key': 'key1', 'Value': 'value2'},  # Should not remove as doesnt match
            {'Key': 'key2'},  # Single key removal
            {'Key': 'key3', 'Value': 'value3'},  # Exact match removal
            {'Key': 'key4'}  # Partial match removal
        ]
    )

    resp = client.list_tags_for_certificate(CertificateArn=arn)
    tags = {item['Key']: item.get('Value', '__NONE__') for item in resp['Tags']}

    for key in ('key2', 'key3', 'key4'):
        tags.should_not.contain(key)

    tags.should.contain('key1')


@mock_acm
def test_remove_tags_from_invalid_certificate():
    client = boto3.client('acm', region_name='eu-central-1')

    try:
        client.remove_tags_from_certificate(
            CertificateArn=BAD_ARN,
            Tags=[
                {'Key': 'key1', 'Value': 'value1'},
                {'Key': 'key2'},
            ]
        )
    except ClientError as err:
        err.response['Error']['Code'].should.equal('ResourceNotFoundException')
    else:
        raise RuntimeError('Should of raised ResourceNotFoundException')


@mock_acm
def test_resend_validation_email():
    client = boto3.client('acm', region_name='eu-central-1')
    arn = _import_cert(client)

    client.resend_validation_email(
        CertificateArn=arn,
        Domain='*.moto.com',
        ValidationDomain='NOTUSEDYET'
    )
    # Returns nothing, boto would raise Exceptions otherwise


@mock_acm
def test_resend_validation_email_invalid():
    client = boto3.client('acm', region_name='eu-central-1')
    arn = _import_cert(client)

    try:
        client.resend_validation_email(
            CertificateArn=arn,
            Domain='no-match.moto.com',
            ValidationDomain='NOTUSEDYET'
        )
    except ClientError as err:
        err.response['Error']['Code'].should.equal('InvalidDomainValidationOptionsException')
    else:
        raise RuntimeError('Should of raised InvalidDomainValidationOptionsException')

    try:
        client.resend_validation_email(
            CertificateArn=BAD_ARN,
            Domain='no-match.moto.com',
            ValidationDomain='NOTUSEDYET'
        )
    except ClientError as err:
        err.response['Error']['Code'].should.equal('ResourceNotFoundException')
    else:
        raise RuntimeError('Should of raised ResourceNotFoundException')


@mock_acm
def test_request_certificate():
    client = boto3.client('acm', region_name='eu-central-1')

    token = str(uuid.uuid4())

    resp = client.request_certificate(
        DomainName='google.com',
        IdempotencyToken=token,
        SubjectAlternativeNames=['google.com', 'www.google.com', 'mail.google.com'],
    )
    resp.should.contain('CertificateArn')
    arn = resp['CertificateArn']

    resp = client.request_certificate(
        DomainName='google.com',
        IdempotencyToken=token,
        SubjectAlternativeNames=['google.com', 'www.google.com', 'mail.google.com'],
    )
    resp['CertificateArn'].should.equal(arn)


@mock_acm
def test_request_certificate_no_san():
    client = boto3.client('acm', region_name='eu-central-1')

    resp = client.request_certificate(
        DomainName='google.com'
    )
    resp.should.contain('CertificateArn')

    resp2 = client.describe_certificate(
        CertificateArn=resp['CertificateArn']
    )
    resp2.should.contain('Certificate')

# # Also tests the SAN code
# # requires Pull: https://github.com/spulec/freezegun/pull/210
# @freeze_time("2012-01-01 12:00:00", as_arg=True)
# @mock_acm
# def test_request_certificate(frozen_time):
#     # After requesting a certificate, it should then auto-validate after 1 minute
#     # Some sneaky programming for that ;-)
#     client = boto3.client('acm', region_name='eu-central-1')
#
#     resp = client.request_certificate(
#         DomainName='google.com',
#         SubjectAlternativeNames=['google.com', 'www.google.com', 'mail.google.com'],
#     )
#     arn = resp['CertificateArn']
#
#     resp = client.describe_certificate(CertificateArn=arn)
#     resp['Certificate']['CertificateArn'].should.equal(arn)
#     resp['Certificate']['DomainName'].should.equal('google.com')
#     resp['Certificate']['Issuer'].should.equal('Amazon')
#     resp['Certificate']['KeyAlgorithm'].should.equal('RSA_2048')
#     resp['Certificate']['Status'].should.equal('PENDING_VALIDATION')
#     resp['Certificate']['Type'].should.equal('AMAZON_ISSUED')
#     len(resp['Certificate']['SubjectAlternativeNames']).should.equal(3)
#
#     # Move time
#     frozen_time.move_to('2012-01-01 12:02:00')
#     resp = client.describe_certificate(CertificateArn=arn)
#     resp['Certificate']['CertificateArn'].should.equal(arn)
#     resp['Certificate']['Status'].should.equal('ISSUED')
#
#
# # requires Pull: https://github.com/spulec/freezegun/pull/210
# @freeze_time("2012-01-01 12:00:00", as_arg=True)
# @mock_acm
# def test_request_certificate(frozen_time):
#     # After requesting a certificate, it should then auto-validate after 1 minute
#     # Some sneaky programming for that ;-)
#     client = boto3.client('acm', region_name='eu-central-1')
#
#     resp = client.request_certificate(
#         IdempotencyToken='test_token',
#         DomainName='google.com',
#         SubjectAlternativeNames=['google.com', 'www.google.com', 'mail.google.com'],
#     )
#     original_arn = resp['CertificateArn']
#
#     # Should be able to request a certificate multiple times in an hour
#     # after that it makes a new one
#     for time_intervals in ('2012-01-01 12:15:00', '2012-01-01 12:30:00', '2012-01-01 12:45:00'):
#         frozen_time.move_to(time_intervals)
#         resp = client.request_certificate(
#             IdempotencyToken='test_token',
#             DomainName='google.com',
#             SubjectAlternativeNames=['google.com', 'www.google.com', 'mail.google.com'],
#         )
#         arn = resp['CertificateArn']
#         arn.should.equal(original_arn)
#
#     # Move time
#     frozen_time.move_to('2012-01-01 13:01:00')
#     resp = client.request_certificate(
#         IdempotencyToken='test_token',
#         DomainName='google.com',
#         SubjectAlternativeNames=['google.com', 'www.google.com', 'mail.google.com'],
#     )
#     arn = resp['CertificateArn']
#     arn.should_not.equal(original_arn)



from __future__ import unicode_literals
from .responses import MotoAPIResponse

url_bases = [
    "https?://motoapi.amazonaws.com"
]

response_instance = MotoAPIResponse()

url_paths = {
    '{0}/moto-api/$': response_instance.dashboard,
    '{0}/moto-api/data.json': response_instance.model_data,
    '{0}/moto-api/reset': response_instance.reset_response,
}



from __future__ import unicode_literals
import time
from datetime import datetime

import boto
import pytz
from boto.emr.bootstrap_action import BootstrapAction
from boto.emr.instance_group import InstanceGroup
from boto.emr.step import StreamingStep

import six
import sure  # noqa

from moto import mock_emr_deprecated
from tests.helpers import requires_boto_gte


run_jobflow_args = dict(
    job_flow_role='EMR_EC2_DefaultRole',
    keep_alive=True,
    log_uri='s3://some_bucket/jobflow_logs',
    master_instance_type='c1.medium',
    name='My jobflow',
    num_instances=2,
    service_role='EMR_DefaultRole',
    slave_instance_type='c1.medium',
)


input_instance_groups = [
    InstanceGroup(1, 'MASTER', 'c1.medium', 'ON_DEMAND', 'master'),
    InstanceGroup(3, 'CORE', 'c1.medium', 'ON_DEMAND', 'core'),
    InstanceGroup(6, 'TASK', 'c1.large', 'SPOT', 'task-1', '0.07'),
    InstanceGroup(10, 'TASK', 'c1.xlarge', 'SPOT', 'task-2', '0.05'),
]


@mock_emr_deprecated
def test_describe_cluster():
    conn = boto.connect_emr()
    args = run_jobflow_args.copy()
    args.update(dict(
        api_params={
            'Applications.member.1.Name': 'Spark',
            'Applications.member.1.Version': '2.4.2',
            'Configurations.member.1.Classification': 'yarn-site',
            'Configurations.member.1.Properties.entry.1.key': 'someproperty',
            'Configurations.member.1.Properties.entry.1.value': 'somevalue',
            'Configurations.member.1.Properties.entry.2.key': 'someotherproperty',
            'Configurations.member.1.Properties.entry.2.value': 'someothervalue',
            'Instances.EmrManagedMasterSecurityGroup': 'master-security-group',
            'Instances.Ec2SubnetId': 'subnet-8be41cec',
        },
        availability_zone='us-east-2b',
        ec2_keyname='mykey',
        job_flow_role='EMR_EC2_DefaultRole',
        keep_alive=False,
        log_uri='s3://some_bucket/jobflow_logs',
        name='My jobflow',
        service_role='EMR_DefaultRole',
        visible_to_all_users=True,
    ))
    cluster_id = conn.run_jobflow(**args)
    input_tags = {'tag1': 'val1', 'tag2': 'val2'}
    conn.add_tags(cluster_id, input_tags)

    cluster = conn.describe_cluster(cluster_id)
    cluster.applications[0].name.should.equal('Spark')
    cluster.applications[0].version.should.equal('2.4.2')
    cluster.autoterminate.should.equal('true')

    # configurations appear not be supplied as attributes?

    attrs = cluster.ec2instanceattributes
    # AdditionalMasterSecurityGroups
    # AdditionalSlaveSecurityGroups
    attrs.ec2availabilityzone.should.equal(args['availability_zone'])
    attrs.ec2keyname.should.equal(args['ec2_keyname'])
    attrs.ec2subnetid.should.equal(args['api_params']['Instances.Ec2SubnetId'])
    # EmrManagedMasterSecurityGroups
    # EmrManagedSlaveSecurityGroups
    attrs.iaminstanceprofile.should.equal(args['job_flow_role'])
    # ServiceAccessSecurityGroup

    cluster.id.should.equal(cluster_id)
    cluster.loguri.should.equal(args['log_uri'])
    cluster.masterpublicdnsname.should.be.a(six.string_types)
    cluster.name.should.equal(args['name'])
    int(cluster.normalizedinstancehours).should.equal(0)
    # cluster.release_label
    cluster.shouldnt.have.property('requestedamiversion')
    cluster.runningamiversion.should.equal('1.0.0')
    # cluster.securityconfiguration
    cluster.servicerole.should.equal(args['service_role'])

    cluster.status.state.should.equal('TERMINATED')
    cluster.status.statechangereason.message.should.be.a(six.string_types)
    cluster.status.statechangereason.code.should.be.a(six.string_types)
    cluster.status.timeline.creationdatetime.should.be.a(six.string_types)
    # cluster.status.timeline.enddatetime.should.be.a(six.string_types)
    # cluster.status.timeline.readydatetime.should.be.a(six.string_types)

    dict((item.key, item.value)
         for item in cluster.tags).should.equal(input_tags)

    cluster.terminationprotected.should.equal('false')
    cluster.visibletoallusers.should.equal('true')


@mock_emr_deprecated
def test_describe_jobflows():
    conn = boto.connect_emr()
    args = run_jobflow_args.copy()
    expected = {}

    for idx in range(4):
        cluster_name = 'cluster' + str(idx)
        args['name'] = cluster_name
        cluster_id = conn.run_jobflow(**args)
        expected[cluster_id] = {
            'id': cluster_id,
            'name': cluster_name,
            'state': 'WAITING'
        }

    # need sleep since it appears the timestamp is always rounded to
    # the nearest second internally
    time.sleep(1)
    timestamp = datetime.now(pytz.utc)
    time.sleep(1)

    for idx in range(4, 6):
        cluster_name = 'cluster' + str(idx)
        args['name'] = cluster_name
        cluster_id = conn.run_jobflow(**args)
        conn.terminate_jobflow(cluster_id)
        expected[cluster_id] = {
            'id': cluster_id,
            'name': cluster_name,
            'state': 'TERMINATED'
        }
    jobs = conn.describe_jobflows()
    jobs.should.have.length_of(6)

    for cluster_id, y in expected.items():
        resp = conn.describe_jobflows(jobflow_ids=[cluster_id])
        resp.should.have.length_of(1)
        resp[0].jobflowid.should.equal(cluster_id)

    resp = conn.describe_jobflows(states=['WAITING'])
    resp.should.have.length_of(4)
    for x in resp:
        x.state.should.equal('WAITING')

    resp = conn.describe_jobflows(created_before=timestamp)
    resp.should.have.length_of(4)

    resp = conn.describe_jobflows(created_after=timestamp)
    resp.should.have.length_of(2)


@mock_emr_deprecated
def test_describe_jobflow():
    conn = boto.connect_emr()
    args = run_jobflow_args.copy()
    args.update(dict(
        ami_version='3.8.1',
        api_params={
            #'Applications.member.1.Name': 'Spark',
            #'Applications.member.1.Version': '2.4.2',
            #'Configurations.member.1.Classification': 'yarn-site',
            #'Configurations.member.1.Properties.entry.1.key': 'someproperty',
            #'Configurations.member.1.Properties.entry.1.value': 'somevalue',
            #'Instances.EmrManagedMasterSecurityGroup': 'master-security-group',
            'Instances.Ec2SubnetId': 'subnet-8be41cec',
        },
        ec2_keyname='mykey',
        hadoop_version='2.4.0',

        name='My jobflow',
        log_uri='s3://some_bucket/jobflow_logs',
        keep_alive=True,
        master_instance_type='c1.medium',
        slave_instance_type='c1.medium',
        num_instances=2,

        availability_zone='us-west-2b',

        job_flow_role='EMR_EC2_DefaultRole',
        service_role='EMR_DefaultRole',
        visible_to_all_users=True,
    ))

    cluster_id = conn.run_jobflow(**args)
    jf = conn.describe_jobflow(cluster_id)
    jf.amiversion.should.equal(args['ami_version'])
    jf.bootstrapactions.should.equal(None)
    jf.creationdatetime.should.be.a(six.string_types)
    jf.should.have.property('laststatechangereason')
    jf.readydatetime.should.be.a(six.string_types)
    jf.startdatetime.should.be.a(six.string_types)
    jf.state.should.equal('WAITING')

    jf.ec2keyname.should.equal(args['ec2_keyname'])
    # Ec2SubnetId
    jf.hadoopversion.should.equal(args['hadoop_version'])
    int(jf.instancecount).should.equal(2)

    for ig in jf.instancegroups:
        ig.creationdatetime.should.be.a(six.string_types)
        # ig.enddatetime.should.be.a(six.string_types)
        ig.should.have.property('instancegroupid').being.a(six.string_types)
        int(ig.instancerequestcount).should.equal(1)
        ig.instancerole.should.be.within(['MASTER', 'CORE'])
        int(ig.instancerunningcount).should.equal(1)
        ig.instancetype.should.equal('c1.medium')
        ig.laststatechangereason.should.be.a(six.string_types)
        ig.market.should.equal('ON_DEMAND')
        ig.name.should.be.a(six.string_types)
        ig.readydatetime.should.be.a(six.string_types)
        ig.startdatetime.should.be.a(six.string_types)
        ig.state.should.equal('RUNNING')

    jf.keepjobflowalivewhennosteps.should.equal('true')
    jf.masterinstanceid.should.be.a(six.string_types)
    jf.masterinstancetype.should.equal(args['master_instance_type'])
    jf.masterpublicdnsname.should.be.a(six.string_types)
    int(jf.normalizedinstancehours).should.equal(0)
    jf.availabilityzone.should.equal(args['availability_zone'])
    jf.slaveinstancetype.should.equal(args['slave_instance_type'])
    jf.terminationprotected.should.equal('false')

    jf.jobflowid.should.equal(cluster_id)
    # jf.jobflowrole.should.equal(args['job_flow_role'])
    jf.loguri.should.equal(args['log_uri'])
    jf.name.should.equal(args['name'])
    # jf.servicerole.should.equal(args['service_role'])

    jf.steps.should.have.length_of(0)

    list(i.value for i in jf.supported_products).should.equal([])
    jf.visibletoallusers.should.equal('true')


@mock_emr_deprecated
def test_list_clusters():
    conn = boto.connect_emr()
    args = run_jobflow_args.copy()
    expected = {}

    for idx in range(40):
        cluster_name = 'jobflow' + str(idx)
        args['name'] = cluster_name
        cluster_id = conn.run_jobflow(**args)
        expected[cluster_id] = {
            'id': cluster_id,
            'name': cluster_name,
            'normalizedinstancehours': '0',
            'state': 'WAITING'
        }

    # need sleep since it appears the timestamp is always rounded to
    # the nearest second internally
    time.sleep(1)
    timestamp = datetime.now(pytz.utc)
    time.sleep(1)

    for idx in range(40, 70):
        cluster_name = 'jobflow' + str(idx)
        args['name'] = cluster_name
        cluster_id = conn.run_jobflow(**args)
        conn.terminate_jobflow(cluster_id)
        expected[cluster_id] = {
            'id': cluster_id,
            'name': cluster_name,
            'normalizedinstancehours': '0',
            'state': 'TERMINATED'
        }

    args = {}
    while 1:
        resp = conn.list_clusters(**args)
        clusters = resp.clusters
        len(clusters).should.be.lower_than_or_equal_to(50)
        for x in clusters:
            y = expected[x.id]
            x.id.should.equal(y['id'])
            x.name.should.equal(y['name'])
            x.normalizedinstancehours.should.equal(
                y['normalizedinstancehours'])
            x.status.state.should.equal(y['state'])
            x.status.timeline.creationdatetime.should.be.a(six.string_types)
            if y['state'] == 'TERMINATED':
                x.status.timeline.enddatetime.should.be.a(six.string_types)
            else:
                x.status.timeline.shouldnt.have.property('enddatetime')
            x.status.timeline.readydatetime.should.be.a(six.string_types)
        if not hasattr(resp, 'marker'):
            break
        args = {'marker': resp.marker}

    resp = conn.list_clusters(cluster_states=['TERMINATED'])
    resp.clusters.should.have.length_of(30)
    for x in resp.clusters:
        x.status.state.should.equal('TERMINATED')

    resp = conn.list_clusters(created_before=timestamp)
    resp.clusters.should.have.length_of(40)

    resp = conn.list_clusters(created_after=timestamp)
    resp.clusters.should.have.length_of(30)


@mock_emr_deprecated
def test_run_jobflow():
    conn = boto.connect_emr()
    args = run_jobflow_args.copy()
    job_id = conn.run_jobflow(**args)
    job_flow = conn.describe_jobflow(job_id)
    job_flow.state.should.equal('WAITING')
    job_flow.jobflowid.should.equal(job_id)
    job_flow.name.should.equal(args['name'])
    job_flow.masterinstancetype.should.equal(args['master_instance_type'])
    job_flow.slaveinstancetype.should.equal(args['slave_instance_type'])
    job_flow.loguri.should.equal(args['log_uri'])
    job_flow.visibletoallusers.should.equal('false')
    int(job_flow.normalizedinstancehours).should.equal(0)
    job_flow.steps.should.have.length_of(0)


@mock_emr_deprecated
def test_run_jobflow_in_multiple_regions():
    regions = {}
    for region in ['us-east-1', 'eu-west-1']:
        conn = boto.emr.connect_to_region(region)
        args = run_jobflow_args.copy()
        args['name'] = region
        cluster_id = conn.run_jobflow(**args)
        regions[region] = {'conn': conn, 'cluster_id': cluster_id}

    for region in regions.keys():
        conn = regions[region]['conn']
        jf = conn.describe_jobflow(regions[region]['cluster_id'])
        jf.name.should.equal(region)


@requires_boto_gte("2.8")
@mock_emr_deprecated
def test_run_jobflow_with_new_params():
    # Test that run_jobflow works with newer params
    conn = boto.connect_emr()
    conn.run_jobflow(**run_jobflow_args)


@requires_boto_gte("2.8")
@mock_emr_deprecated
def test_run_jobflow_with_visible_to_all_users():
    conn = boto.connect_emr()
    for expected in (True, False):
        job_id = conn.run_jobflow(
            visible_to_all_users=expected,
            **run_jobflow_args
        )
        job_flow = conn.describe_jobflow(job_id)
        job_flow.visibletoallusers.should.equal(str(expected).lower())


@requires_boto_gte("2.8")
@mock_emr_deprecated
def test_run_jobflow_with_instance_groups():
    input_groups = dict((g.name, g) for g in input_instance_groups)
    conn = boto.connect_emr()
    job_id = conn.run_jobflow(instance_groups=input_instance_groups,
                              **run_jobflow_args)
    job_flow = conn.describe_jobflow(job_id)
    int(job_flow.instancecount).should.equal(
        sum(g.num_instances for g in input_instance_groups))
    for instance_group in job_flow.instancegroups:
        expected = input_groups[instance_group.name]
        instance_group.should.have.property('instancegroupid')
        int(instance_group.instancerunningcount).should.equal(
            expected.num_instances)
        instance_group.instancerole.should.equal(expected.role)
        instance_group.instancetype.should.equal(expected.type)
        instance_group.market.should.equal(expected.market)
        if hasattr(expected, 'bidprice'):
            instance_group.bidprice.should.equal(expected.bidprice)


@requires_boto_gte("2.8")
@mock_emr_deprecated
def test_set_termination_protection():
    conn = boto.connect_emr()
    job_id = conn.run_jobflow(**run_jobflow_args)
    job_flow = conn.describe_jobflow(job_id)
    job_flow.terminationprotected.should.equal('false')

    conn.set_termination_protection(job_id, True)
    job_flow = conn.describe_jobflow(job_id)
    job_flow.terminationprotected.should.equal('true')

    conn.set_termination_protection(job_id, False)
    job_flow = conn.describe_jobflow(job_id)
    job_flow.terminationprotected.should.equal('false')


@requires_boto_gte("2.8")
@mock_emr_deprecated
def test_set_visible_to_all_users():
    conn = boto.connect_emr()
    args = run_jobflow_args.copy()
    args['visible_to_all_users'] = False
    job_id = conn.run_jobflow(**args)
    job_flow = conn.describe_jobflow(job_id)
    job_flow.visibletoallusers.should.equal('false')

    conn.set_visible_to_all_users(job_id, True)
    job_flow = conn.describe_jobflow(job_id)
    job_flow.visibletoallusers.should.equal('true')

    conn.set_visible_to_all_users(job_id, False)
    job_flow = conn.describe_jobflow(job_id)
    job_flow.visibletoallusers.should.equal('false')


@mock_emr_deprecated
def test_terminate_jobflow():
    conn = boto.connect_emr()
    job_id = conn.run_jobflow(**run_jobflow_args)
    flow = conn.describe_jobflows()[0]
    flow.state.should.equal('WAITING')

    conn.terminate_jobflow(job_id)
    flow = conn.describe_jobflows()[0]
    flow.state.should.equal('TERMINATED')


# testing multiple end points for each feature

@mock_emr_deprecated
def test_bootstrap_actions():
    bootstrap_actions = [
        BootstrapAction(
            name='bs1',
            path='path/to/script',
            bootstrap_action_args=['arg1', 'arg2&arg3']),
        BootstrapAction(
            name='bs2',
            path='path/to/anotherscript',
            bootstrap_action_args=[])
    ]

    conn = boto.connect_emr()
    cluster_id = conn.run_jobflow(
        bootstrap_actions=bootstrap_actions,
        **run_jobflow_args
    )

    jf = conn.describe_jobflow(cluster_id)
    for x, y in zip(jf.bootstrapactions, bootstrap_actions):
        x.name.should.equal(y.name)
        x.path.should.equal(y.path)
        list(o.value for o in x.args).should.equal(y.args())

    resp = conn.list_bootstrap_actions(cluster_id)
    for i, y in enumerate(bootstrap_actions):
        x = resp.actions[i]
        x.name.should.equal(y.name)
        x.scriptpath.should.equal(y.path)
        list(arg.value for arg in x.args).should.equal(y.args())


@mock_emr_deprecated
def test_instance_groups():
    input_groups = dict((g.name, g) for g in input_instance_groups)

    conn = boto.connect_emr()
    args = run_jobflow_args.copy()
    for key in ['master_instance_type', 'slave_instance_type', 'num_instances']:
        del args[key]
    args['instance_groups'] = input_instance_groups[:2]
    job_id = conn.run_jobflow(**args)

    jf = conn.describe_jobflow(job_id)
    base_instance_count = int(jf.instancecount)

    conn.add_instance_groups(job_id, input_instance_groups[2:])

    jf = conn.describe_jobflow(job_id)
    int(jf.instancecount).should.equal(
        sum(g.num_instances for g in input_instance_groups))
    for x in jf.instancegroups:
        y = input_groups[x.name]
        if hasattr(y, 'bidprice'):
            x.bidprice.should.equal(y.bidprice)
        x.creationdatetime.should.be.a(six.string_types)
        # x.enddatetime.should.be.a(six.string_types)
        x.should.have.property('instancegroupid')
        int(x.instancerequestcount).should.equal(y.num_instances)
        x.instancerole.should.equal(y.role)
        int(x.instancerunningcount).should.equal(y.num_instances)
        x.instancetype.should.equal(y.type)
        x.laststatechangereason.should.be.a(six.string_types)
        x.market.should.equal(y.market)
        x.name.should.be.a(six.string_types)
        x.readydatetime.should.be.a(six.string_types)
        x.startdatetime.should.be.a(six.string_types)
        x.state.should.equal('RUNNING')

    for x in conn.list_instance_groups(job_id).instancegroups:
        y = input_groups[x.name]
        if hasattr(y, 'bidprice'):
            x.bidprice.should.equal(y.bidprice)
        # Configurations
        # EbsBlockDevices
        # EbsOptimized
        x.should.have.property('id')
        x.instancegrouptype.should.equal(y.role)
        x.instancetype.should.equal(y.type)
        x.market.should.equal(y.market)
        x.name.should.equal(y.name)
        int(x.requestedinstancecount).should.equal(y.num_instances)
        int(x.runninginstancecount).should.equal(y.num_instances)
        # ShrinkPolicy
        x.status.state.should.equal('RUNNING')
        x.status.statechangereason.code.should.be.a(six.string_types)
        x.status.statechangereason.message.should.be.a(six.string_types)
        x.status.timeline.creationdatetime.should.be.a(six.string_types)
        # x.status.timeline.enddatetime.should.be.a(six.string_types)
        x.status.timeline.readydatetime.should.be.a(six.string_types)

    igs = dict((g.name, g) for g in jf.instancegroups)

    conn.modify_instance_groups(
        [igs['task-1'].instancegroupid, igs['task-2'].instancegroupid],
        [2, 3])
    jf = conn.describe_jobflow(job_id)
    int(jf.instancecount).should.equal(base_instance_count + 5)
    igs = dict((g.name, g) for g in jf.instancegroups)
    int(igs['task-1'].instancerunningcount).should.equal(2)
    int(igs['task-2'].instancerunningcount).should.equal(3)


@mock_emr_deprecated
def test_steps():
    input_steps = [
        StreamingStep(
            name='My wordcount example',
            mapper='s3n://elasticmapreduce/samples/wordcount/wordSplitter.py',
            reducer='aggregate',
            input='s3n://elasticmapreduce/samples/wordcount/input',
            output='s3n://output_bucket/output/wordcount_output'),
        StreamingStep(
            name='My wordcount example & co.',
            mapper='s3n://elasticmapreduce/samples/wordcount/wordSplitter2.py',
            reducer='aggregate',
            input='s3n://elasticmapreduce/samples/wordcount/input2',
            output='s3n://output_bucket/output/wordcount_output2')
    ]

    # TODO: implementation and test for cancel_steps

    conn = boto.connect_emr()
    cluster_id = conn.run_jobflow(
        steps=[input_steps[0]],
        **run_jobflow_args)

    jf = conn.describe_jobflow(cluster_id)
    jf.steps.should.have.length_of(1)

    conn.add_jobflow_steps(cluster_id, [input_steps[1]])

    jf = conn.describe_jobflow(cluster_id)
    jf.steps.should.have.length_of(2)
    for step in jf.steps:
        step.actiononfailure.should.equal('TERMINATE_JOB_FLOW')
        list(arg.value for arg in step.args).should.have.length_of(8)
        step.creationdatetime.should.be.a(six.string_types)
        # step.enddatetime.should.be.a(six.string_types)
        step.jar.should.equal(
            '/home/hadoop/contrib/streaming/hadoop-streaming.jar')
        step.laststatechangereason.should.be.a(six.string_types)
        step.mainclass.should.equal('')
        step.name.should.be.a(six.string_types)
        # step.readydatetime.should.be.a(six.string_types)
        # step.startdatetime.should.be.a(six.string_types)
        step.state.should.be.within(['STARTING', 'PENDING'])

    expected = dict((s.name, s) for s in input_steps)

    steps = conn.list_steps(cluster_id).steps
    for x in steps:
        y = expected[x.name]
        # actiononfailure
        list(arg.value for arg in x.config.args).should.equal([
            '-mapper', y.mapper,
            '-reducer', y.reducer,
            '-input', y.input,
            '-output', y.output,
        ])
        x.config.jar.should.equal(
            '/home/hadoop/contrib/streaming/hadoop-streaming.jar')
        x.config.mainclass.should.equal('')
        # properties
        x.should.have.property('id').should.be.a(six.string_types)
        x.name.should.equal(y.name)
        x.status.state.should.be.within(['STARTING', 'PENDING'])
        # x.status.statechangereason
        x.status.timeline.creationdatetime.should.be.a(six.string_types)
        # x.status.timeline.enddatetime.should.be.a(six.string_types)
        # x.status.timeline.startdatetime.should.be.a(six.string_types)

        x = conn.describe_step(cluster_id, x.id)
        list(arg.value for arg in x.config.args).should.equal([
            '-mapper', y.mapper,
            '-reducer', y.reducer,
            '-input', y.input,
            '-output', y.output,
        ])
        x.config.jar.should.equal(
            '/home/hadoop/contrib/streaming/hadoop-streaming.jar')
        x.config.mainclass.should.equal('')
        # properties
        x.should.have.property('id').should.be.a(six.string_types)
        x.name.should.equal(y.name)
        x.status.state.should.be.within(['STARTING', 'PENDING'])
        # x.status.statechangereason
        x.status.timeline.creationdatetime.should.be.a(six.string_types)
        # x.status.timeline.enddatetime.should.be.a(six.string_types)
        # x.status.timeline.startdatetime.should.be.a(six.string_types)

    @requires_boto_gte('2.39')
    def test_list_steps_with_states():
        # boto's list_steps prior to 2.39 has a bug that ignores
        # step_states argument.
        steps = conn.list_steps(cluster_id).steps
        step_id = steps[0].id
        steps = conn.list_steps(cluster_id, step_states=['STARTING']).steps
        steps.should.have.length_of(1)
        steps[0].id.should.equal(step_id)
    test_list_steps_with_states()


@mock_emr_deprecated
def test_tags():
    input_tags = {"tag1": "val1", "tag2": "val2"}

    conn = boto.connect_emr()
    cluster_id = conn.run_jobflow(**run_jobflow_args)

    conn.add_tags(cluster_id, input_tags)
    cluster = conn.describe_cluster(cluster_id)
    cluster.tags.should.have.length_of(2)
    dict((t.key, t.value) for t in cluster.tags).should.equal(input_tags)

    conn.remove_tags(cluster_id, list(input_tags.keys()))
    cluster = conn.describe_cluster(cluster_id)
    cluster.tags.should.have.length_of(0)



from __future__ import unicode_literals
import boto
import boto.ec2
import boto3
import sure  # noqa

from moto import mock_ec2, mock_ec2_deprecated


@mock_ec2_deprecated
def test_describe_regions():
    conn = boto.connect_ec2('the_key', 'the_secret')
    regions = conn.get_all_regions()
    regions.should.have.length_of(16)
    for region in regions:
        region.endpoint.should.contain(region.name)


@mock_ec2_deprecated
def test_availability_zones():
    conn = boto.connect_ec2('the_key', 'the_secret')
    regions = conn.get_all_regions()
    for region in regions:
        conn = boto.ec2.connect_to_region(region.name)
        if conn is None:
            continue
        for zone in conn.get_all_zones():
            zone.name.should.contain(region.name)


@mock_ec2
def test_boto3_describe_regions():
    ec2 = boto3.client('ec2', 'us-east-1')
    resp = ec2.describe_regions()
    resp['Regions'].should.have.length_of(16)
    for rec in resp['Regions']:
        rec['Endpoint'].should.contain(rec['RegionName'])

    test_region = 'us-east-1' 
    resp = ec2.describe_regions(RegionNames=[test_region])
    resp['Regions'].should.have.length_of(1)
    resp['Regions'][0].should.have.key('RegionName').which.should.equal(test_region)


@mock_ec2
def test_boto3_availability_zones():
    ec2 = boto3.client('ec2', 'us-east-1')
    resp = ec2.describe_regions()
    regions = [r['RegionName'] for r in resp['Regions']]
    for region in regions:
        conn = boto3.client('ec2', region)
        resp = conn.describe_availability_zones()
        for rec in resp['AvailabilityZones']:
            rec['ZoneName'].should.contain(region)



from __future__ import unicode_literals
from .models import batch_backends
from ..core.models import base_decorator

batch_backend = batch_backends['us-east-1']
mock_batch = base_decorator(batch_backends)



from __future__ import unicode_literals

import json

try:
    from urllib import unquote
except ImportError:
    from urllib.parse import unquote

from moto.core.utils import amz_crc32, amzn_request_id, path_url
from moto.core.responses import BaseResponse
from .models import lambda_backends


class LambdaResponse(BaseResponse):
    @property
    def json_body(self):
        """
        :return: JSON
        :rtype: dict
        """
        return json.loads(self.body)

    @property
    def lambda_backend(self):
        """
        Get backend
        :return: Lambda Backend
        :rtype: moto.awslambda.models.LambdaBackend
        """
        return lambda_backends[self.region]

    def root(self, request, full_url, headers):
        self.setup_class(request, full_url, headers)
        if request.method == 'GET':
            return self._list_functions(request, full_url, headers)
        elif request.method == 'POST':
            return self._create_function(request, full_url, headers)
        else:
            raise ValueError("Cannot handle request")

    def function(self, request, full_url, headers):
        self.setup_class(request, full_url, headers)
        if request.method == 'GET':
            return self._get_function(request, full_url, headers)
        elif request.method == 'DELETE':
            return self._delete_function(request, full_url, headers)
        else:
            raise ValueError("Cannot handle request")

    def versions(self, request, full_url, headers):
        self.setup_class(request, full_url, headers)
        if request.method == 'GET':
            # This is ListVersionByFunction

            path = request.path if hasattr(request, 'path') else path_url(request.url)
            function_name = path.split('/')[-2]
            return self._list_versions_by_function(function_name)

        elif request.method == 'POST':
            return self._publish_function(request, full_url, headers)
        else:
            raise ValueError("Cannot handle request")

    @amz_crc32
    @amzn_request_id
    def invoke(self, request, full_url, headers):
        self.setup_class(request, full_url, headers)
        if request.method == 'POST':
            return self._invoke(request, full_url)
        else:
            raise ValueError("Cannot handle request")

    @amz_crc32
    @amzn_request_id
    def invoke_async(self, request, full_url, headers):
        self.setup_class(request, full_url, headers)
        if request.method == 'POST':
            return self._invoke_async(request, full_url)
        else:
            raise ValueError("Cannot handle request")

    def tag(self, request, full_url, headers):
        self.setup_class(request, full_url, headers)
        if request.method == 'GET':
            return self._list_tags(request, full_url)
        elif request.method == 'POST':
            return self._tag_resource(request, full_url)
        elif request.method == 'DELETE':
            return self._untag_resource(request, full_url)
        else:
            raise ValueError("Cannot handle {0} request".format(request.method))

    def policy(self, request, full_url, headers):
        if request.method == 'GET':
            return self._get_policy(request, full_url, headers)
        if request.method == 'POST':
            return self._add_policy(request, full_url, headers)

    def _add_policy(self, request, full_url, headers):
        path = request.path if hasattr(request, 'path') else path_url(request.url)
        function_name = path.split('/')[-2]
        if self.lambda_backend.get_function(function_name):
            policy = request.body.decode('utf8')
            self.lambda_backend.add_policy(function_name, policy)
            return 200, {}, json.dumps(dict(Statement=policy))
        else:
            return 404, {}, "{}"

    def _get_policy(self, request, full_url, headers):
        path = request.path if hasattr(request, 'path') else path_url(request.url)
        function_name = path.split('/')[-2]
        if self.lambda_backend.get_function(function_name):
            lambda_function = self.lambda_backend.get_function(function_name)
            return 200, {}, json.dumps(dict(Policy="{\"Statement\":[" + lambda_function.policy + "]}"))
        else:
            return 404, {}, "{}"

    def _invoke(self, request, full_url):
        response_headers = {}

        function_name = self.path.rsplit('/', 2)[-2]
        qualifier = self._get_param('qualifier')

        fn = self.lambda_backend.get_function(function_name, qualifier)
        if fn:
            payload = fn.invoke(self.body, self.headers, response_headers)
            response_headers['Content-Length'] = str(len(payload))
            return 202, response_headers, payload
        else:
            return 404, response_headers, "{}"

    def _invoke_async(self, request, full_url):
        response_headers = {}

        function_name = self.path.rsplit('/', 3)[-3]

        fn = self.lambda_backend.get_function(function_name, None)
        if fn:
            payload = fn.invoke(self.body, self.headers, response_headers)
            response_headers['Content-Length'] = str(len(payload))
            return 202, response_headers, payload
        else:
            return 404, response_headers, "{}"

    def _list_functions(self, request, full_url, headers):
        result = {
            'Functions': []
        }

        for fn in self.lambda_backend.list_functions():
            json_data = fn.get_configuration()
            json_data['Version'] = '$LATEST'
            result['Functions'].append(json_data)

        return 200, {}, json.dumps(result)

    def _list_versions_by_function(self, function_name):
        result = {
            'Versions': []
        }

        functions = self.lambda_backend.list_versions_by_function(function_name)
        if functions:
            for fn in functions:
                json_data = fn.get_configuration()
                result['Versions'].append(json_data)

        return 200, {}, json.dumps(result)

    def _create_function(self, request, full_url, headers):
        try:
            fn = self.lambda_backend.create_function(self.json_body)
        except ValueError as e:
            return 400, {}, json.dumps({"Error": {"Code": e.args[0], "Message": e.args[1]}})
        else:
            config = fn.get_configuration()
            return 201, {}, json.dumps(config)

    def _publish_function(self, request, full_url, headers):
        function_name = self.path.rsplit('/', 2)[-2]

        fn = self.lambda_backend.publish_function(function_name)
        if fn:
            config = fn.get_configuration()
            return 201, {}, json.dumps(config)
        else:
            return 404, {}, "{}"

    def _delete_function(self, request, full_url, headers):
        function_name = self.path.rsplit('/', 1)[-1]
        qualifier = self._get_param('Qualifier', None)

        if self.lambda_backend.delete_function(function_name, qualifier):
            return 204, {}, ""
        else:
            return 404, {}, "{}"

    def _get_function(self, request, full_url, headers):
        function_name = self.path.rsplit('/', 1)[-1]
        qualifier = self._get_param('Qualifier', None)

        fn = self.lambda_backend.get_function(function_name, qualifier)

        if fn:
            code = fn.get_code()
            if qualifier is None or qualifier == '$LATEST':
                code['Configuration']['Version'] = '$LATEST'
            if qualifier == '$LATEST':
                code['Configuration']['FunctionArn'] += ':$LATEST'
            return 200, {}, json.dumps(code)
        else:
            return 404, {}, "{}"

    def _get_aws_region(self, full_url):
        region = self.region_regex.search(full_url)
        if region:
            return region.group(1)
        else:
            return self.default_region

    def _list_tags(self, request, full_url):
        function_arn = unquote(self.path.rsplit('/', 1)[-1])

        fn = self.lambda_backend.get_function_by_arn(function_arn)
        if fn:
            return 200, {}, json.dumps({'Tags': fn.tags})
        else:
            return 404, {}, "{}"

    def _tag_resource(self, request, full_url):
        function_arn = unquote(self.path.rsplit('/', 1)[-1])

        if self.lambda_backend.tag_resource(function_arn, self.json_body['Tags']):
            return 200, {}, "{}"
        else:
            return 404, {}, "{}"

    def _untag_resource(self, request, full_url):
        function_arn = unquote(self.path.rsplit('/', 1)[-1])
        tag_keys = self.querystring['tagKeys']

        if self.lambda_backend.untag_resource(function_arn, tag_keys):
            return 204, {}, "{}"
        else:
            return 404, {}, "{}"



from __future__ import unicode_literals

import json
from werkzeug.exceptions import BadRequest


class ResourceNotFoundError(BadRequest):

    def __init__(self, message):
        super(ResourceNotFoundError, self).__init__()
        self.description = json.dumps({
            "message": message,
            '__type': 'ResourceNotFoundException',
        })


class UserNotFoundError(BadRequest):

    def __init__(self, message):
        super(UserNotFoundError, self).__init__()
        self.description = json.dumps({
            "message": message,
            '__type': 'UserNotFoundException',
        })


class GroupExistsException(BadRequest):

    def __init__(self, message):
        super(GroupExistsException, self).__init__()
        self.description = json.dumps({
            "message": message,
            '__type': 'GroupExistsException',
        })


class NotAuthorizedError(BadRequest):

    def __init__(self, message):
        super(NotAuthorizedError, self).__init__()
        self.description = json.dumps({
            "message": message,
            '__type': 'NotAuthorizedException',
        })



#!/usr/bin/env python
from __future__ import unicode_literals
import codecs
import os
import re
import setuptools
from setuptools import setup, find_packages
import sys


# Borrowed from pip at https://github.com/pypa/pip/blob/62c27dee45625e1b63d1e023b0656310f276e050/setup.py#L11-L15
here = os.path.abspath(os.path.dirname(__file__))

def read(*parts):
    # intentionally *not* adding an encoding option to open, See:
    #   https://github.com/pypa/virtualenv/issues/201#issuecomment-3145690
    with codecs.open(os.path.join(here, *parts), 'r') as fp:
        return fp.read()


def get_version():
    version_file = read('moto', '__init__.py')
    version_match = re.search(r'^__version__ = [\'"]([^\'"]*)[\'"]',
                              version_file, re.MULTILINE)
    if version_match:
        return version_match.group(1)
    raise RuntimeError('Unable to find version string.')


install_requires = [
    "Jinja2>=2.10.1",
    "boto>=2.36.0",
    "boto3>=1.9.86",
    "botocore>=1.12.86",
    "cryptography>=2.3.0",
    "datetime",
    "requests>=2.5",
    "xmltodict",
    "six>1.9",
    "werkzeug",
    "PyYAML>=5.1",
    "pytz",
    "python-dateutil<3.0.0,>=2.1",
    "python-jose<4.0.0",
    "mock",
    "docker>=2.5.1",
    "jsondiff==1.1.2",
    "aws-xray-sdk!=0.96,>=0.93",
    "responses>=0.9.0",
    "idna<2.9,>=2.5",
    "cfn-lint",
    "sshpubkeys>=3.1.0,<4.0"
]

extras_require = {
    'server': ['flask'],
}

# https://hynek.me/articles/conditional-python-dependencies/
if int(setuptools.__version__.split(".", 1)[0]) < 18:
    if sys.version_info[0:2] < (3, 3):
        install_requires.append("backports.tempfile")
else:
    extras_require[":python_version<'3.3'"] = ["backports.tempfile"]


setup(
    name='moto',
    version=get_version(),
    description='A library that allows your python tests to easily'
                ' mock out the boto library',
    long_description=read('README.md'),
    long_description_content_type='text/markdown',
    author='Steve Pulec',
    author_email='spulec@gmail.com',
    url='https://github.com/spulec/moto',
    entry_points={
        'console_scripts': [
            'moto_server = moto.server:main',
        ],
    },
    packages=find_packages(exclude=("tests", "tests.*")),
    install_requires=install_requires,
    extras_require=extras_require,
    include_package_data=True,
    license="Apache",
    test_suite="tests",
    classifiers=[
        "Programming Language :: Python :: 2",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.4",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "License :: OSI Approved :: Apache Software License",
        "Topic :: Software Development :: Testing",
    ],
)



from __future__ import unicode_literals

import datetime
import functools
import itertools
import json
import os
import time
import uuid

import boto.cognito.identity
from jose import jws

from moto.compat import OrderedDict
from moto.core import BaseBackend, BaseModel
from .exceptions import GroupExistsException, NotAuthorizedError, ResourceNotFoundError, UserNotFoundError

UserStatus = {
    "FORCE_CHANGE_PASSWORD": "FORCE_CHANGE_PASSWORD",
    "CONFIRMED": "CONFIRMED",
}


def paginate(limit, start_arg="next_token", limit_arg="max_results"):
    """Returns a limited result list, and an offset into list of remaining items

    Takes the next_token, and max_results kwargs given to a function and handles
    the slicing of the results. The kwarg `next_token` is the offset into the
    list to begin slicing from. `max_results` is the size of the result required

    If the max_results is not supplied then the `limit` parameter is used as a
    default

    :param limit_arg: the name of argument in the decorated function that
    controls amount of items returned
    :param start_arg: the name of the argument in the decorated that provides
    the starting offset
    :param limit: A default maximum items to return
    :return: a tuple containing a list of items, and the offset into the list
    """
    default_start = 0

    def outer_wrapper(func):
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            start = int(default_start if kwargs.get(start_arg) is None else kwargs[start_arg])
            lim = int(limit if kwargs.get(limit_arg) is None else kwargs[limit_arg])
            stop = start + lim
            result = func(*args, **kwargs)
            limited_results = list(itertools.islice(result, start, stop))
            next_token = stop if stop < len(result) else None
            return limited_results, next_token
        return wrapper
    return outer_wrapper


class CognitoIdpUserPool(BaseModel):

    def __init__(self, region, name, extended_config):
        self.region = region
        self.id = "{}_{}".format(self.region, str(uuid.uuid4().hex))
        self.name = name
        self.status = None
        self.extended_config = extended_config or {}
        self.creation_date = datetime.datetime.utcnow()
        self.last_modified_date = datetime.datetime.utcnow()

        self.clients = OrderedDict()
        self.identity_providers = OrderedDict()
        self.groups = OrderedDict()
        self.users = OrderedDict()
        self.refresh_tokens = {}
        self.access_tokens = {}
        self.id_tokens = {}

        with open(os.path.join(os.path.dirname(__file__), "resources/jwks-private.json")) as f:
            self.json_web_key = json.loads(f.read())

    def _base_json(self):
        return {
            "Id": self.id,
            "Name": self.name,
            "Status": self.status,
            "CreationDate": time.mktime(self.creation_date.timetuple()),
            "LastModifiedDate": time.mktime(self.last_modified_date.timetuple()),
        }

    def to_json(self, extended=False):
        user_pool_json = self._base_json()
        if extended:
            user_pool_json.update(self.extended_config)
        else:
            user_pool_json["LambdaConfig"] = self.extended_config.get("LambdaConfig") or {}

        return user_pool_json

    def create_jwt(self, client_id, username, expires_in=60 * 60, extra_data={}):
        now = int(time.time())
        payload = {
            "iss": "https://cognito-idp.{}.amazonaws.com/{}".format(self.region, self.id),
            "sub": self.users[username].id,
            "aud": client_id,
            "token_use": "id",
            "auth_time": now,
            "exp": now + expires_in,
        }
        payload.update(extra_data)

        return jws.sign(payload, self.json_web_key, algorithm='RS256'), expires_in

    def create_id_token(self, client_id, username):
        id_token, expires_in = self.create_jwt(client_id, username)
        self.id_tokens[id_token] = (client_id, username)
        return id_token, expires_in

    def create_refresh_token(self, client_id, username):
        refresh_token = str(uuid.uuid4())
        self.refresh_tokens[refresh_token] = (client_id, username)
        return refresh_token

    def create_access_token(self, client_id, username):
        extra_data = self.get_user_extra_data_by_client_id(
            client_id, username
        )
        access_token, expires_in = self.create_jwt(client_id, username,
                                                   extra_data=extra_data)
        self.access_tokens[access_token] = (client_id, username)
        return access_token, expires_in

    def create_tokens_from_refresh_token(self, refresh_token):
        client_id, username = self.refresh_tokens.get(refresh_token)
        if not username:
            raise NotAuthorizedError(refresh_token)

        access_token, expires_in = self.create_access_token(client_id, username)
        id_token, _ = self.create_id_token(client_id, username)
        return access_token, id_token, expires_in

    def get_user_extra_data_by_client_id(self, client_id, username):
        extra_data = {}
        current_client = self.clients.get(client_id, None)
        if current_client:
            for readable_field in current_client.get_readable_fields():
                attribute = list(filter(
                    lambda f: f['Name'] == readable_field,
                    self.users.get(username).attributes
                ))
                if len(attribute) > 0:
                    extra_data.update({
                        attribute[0]['Name']: attribute[0]['Value']
                    })
        return extra_data


class CognitoIdpUserPoolDomain(BaseModel):

    def __init__(self, user_pool_id, domain):
        self.user_pool_id = user_pool_id
        self.domain = domain

    def to_json(self):
        return {
            "UserPoolId": self.user_pool_id,
            "AWSAccountId": str(uuid.uuid4()),
            "CloudFrontDistribution": None,
            "Domain": self.domain,
            "S3Bucket": None,
            "Status": "ACTIVE",
            "Version": None,
        }


class CognitoIdpUserPoolClient(BaseModel):

    def __init__(self, user_pool_id, extended_config):
        self.user_pool_id = user_pool_id
        self.id = str(uuid.uuid4())
        self.secret = str(uuid.uuid4())
        self.extended_config = extended_config or {}

    def _base_json(self):
        return {
            "ClientId": self.id,
            "ClientName": self.extended_config.get("ClientName"),
            "UserPoolId": self.user_pool_id,
        }

    def to_json(self, extended=False):
        user_pool_client_json = self._base_json()
        if extended:
            user_pool_client_json.update(self.extended_config)

        return user_pool_client_json

    def get_readable_fields(self):
        return self.extended_config.get('ReadAttributes', [])


class CognitoIdpIdentityProvider(BaseModel):

    def __init__(self, name, extended_config):
        self.name = name
        self.extended_config = extended_config or {}
        self.creation_date = datetime.datetime.utcnow()
        self.last_modified_date = datetime.datetime.utcnow()

    def _base_json(self):
        return {
            "ProviderName": self.name,
            "ProviderType": self.extended_config.get("ProviderType"),
            "CreationDate": time.mktime(self.creation_date.timetuple()),
            "LastModifiedDate": time.mktime(self.last_modified_date.timetuple()),
        }

    def to_json(self, extended=False):
        identity_provider_json = self._base_json()
        if extended:
            identity_provider_json.update(self.extended_config)

        return identity_provider_json


class CognitoIdpGroup(BaseModel):

    def __init__(self, user_pool_id, group_name, description, role_arn, precedence):
        self.user_pool_id = user_pool_id
        self.group_name = group_name
        self.description = description or ""
        self.role_arn = role_arn
        self.precedence = precedence
        self.last_modified_date = datetime.datetime.now()
        self.creation_date = self.last_modified_date

        # Users who are members of this group.
        # Note that these links are bidirectional.
        self.users = set()

    def to_json(self):
        return {
            "GroupName": self.group_name,
            "UserPoolId": self.user_pool_id,
            "Description": self.description,
            "RoleArn": self.role_arn,
            "Precedence": self.precedence,
            "LastModifiedDate": time.mktime(self.last_modified_date.timetuple()),
            "CreationDate": time.mktime(self.creation_date.timetuple()),
        }


class CognitoIdpUser(BaseModel):

    def __init__(self, user_pool_id, username, password, status, attributes):
        self.id = str(uuid.uuid4())
        self.user_pool_id = user_pool_id
        self.username = username
        self.password = password
        self.status = status
        self.enabled = True
        self.attributes = attributes
        self.create_date = datetime.datetime.utcnow()
        self.last_modified_date = datetime.datetime.utcnow()

        # Groups this user is a member of.
        # Note that these links are bidirectional.
        self.groups = set()

    def _base_json(self):
        return {
            "UserPoolId": self.user_pool_id,
            "Username": self.username,
            "UserStatus": self.status,
            "UserCreateDate": time.mktime(self.create_date.timetuple()),
            "UserLastModifiedDate": time.mktime(self.last_modified_date.timetuple()),
        }

    # list_users brings back "Attributes" while admin_get_user brings back "UserAttributes".
    def to_json(self, extended=False, attributes_key="Attributes"):
        user_json = self._base_json()
        if extended:
            user_json.update(
                {
                    "Enabled": self.enabled,
                    attributes_key: self.attributes,
                    "MFAOptions": []
                }
            )

        return user_json

    def update_attributes(self, new_attributes):

        def flatten_attrs(attrs):
            return {attr['Name']: attr['Value'] for attr in attrs}

        def expand_attrs(attrs):
            return [{'Name': k, 'Value': v} for k, v in attrs.items()]

        flat_attributes = flatten_attrs(self.attributes)
        flat_attributes.update(flatten_attrs(new_attributes))
        self.attributes = expand_attrs(flat_attributes)


class CognitoIdpBackend(BaseBackend):

    def __init__(self, region):
        super(CognitoIdpBackend, self).__init__()
        self.region = region
        self.user_pools = OrderedDict()
        self.user_pool_domains = OrderedDict()
        self.sessions = {}

    def reset(self):
        region = self.region
        self.__dict__ = {}
        self.__init__(region)

    # User pool
    def create_user_pool(self, name, extended_config):
        user_pool = CognitoIdpUserPool(self.region, name, extended_config)
        self.user_pools[user_pool.id] = user_pool
        return user_pool

    @paginate(60)
    def list_user_pools(self, max_results=None, next_token=None):
        return self.user_pools.values()

    def describe_user_pool(self, user_pool_id):
        user_pool = self.user_pools.get(user_pool_id)
        if not user_pool:
            raise ResourceNotFoundError(user_pool_id)

        return user_pool

    def delete_user_pool(self, user_pool_id):
        if user_pool_id not in self.user_pools:
            raise ResourceNotFoundError(user_pool_id)

        del self.user_pools[user_pool_id]

    # User pool domain
    def create_user_pool_domain(self, user_pool_id, domain):
        if user_pool_id not in self.user_pools:
            raise ResourceNotFoundError(user_pool_id)

        user_pool_domain = CognitoIdpUserPoolDomain(user_pool_id, domain)
        self.user_pool_domains[domain] = user_pool_domain
        return user_pool_domain

    def describe_user_pool_domain(self, domain):
        if domain not in self.user_pool_domains:
            return None

        return self.user_pool_domains[domain]

    def delete_user_pool_domain(self, domain):
        if domain not in self.user_pool_domains:
            raise ResourceNotFoundError(domain)

        del self.user_pool_domains[domain]

    # User pool client
    def create_user_pool_client(self, user_pool_id, extended_config):
        user_pool = self.user_pools.get(user_pool_id)
        if not user_pool:
            raise ResourceNotFoundError(user_pool_id)

        user_pool_client = CognitoIdpUserPoolClient(user_pool_id, extended_config)
        user_pool.clients[user_pool_client.id] = user_pool_client
        return user_pool_client

    @paginate(60)
    def list_user_pool_clients(self, user_pool_id, max_results=None, next_token=None):
        user_pool = self.user_pools.get(user_pool_id)
        if not user_pool:
            raise ResourceNotFoundError(user_pool_id)

        return user_pool.clients.values()

    def describe_user_pool_client(self, user_pool_id, client_id):
        user_pool = self.user_pools.get(user_pool_id)
        if not user_pool:
            raise ResourceNotFoundError(user_pool_id)

        client = user_pool.clients.get(client_id)
        if not client:
            raise ResourceNotFoundError(client_id)

        return client

    def update_user_pool_client(self, user_pool_id, client_id, extended_config):
        user_pool = self.user_pools.get(user_pool_id)
        if not user_pool:
            raise ResourceNotFoundError(user_pool_id)

        client = user_pool.clients.get(client_id)
        if not client:
            raise ResourceNotFoundError(client_id)

        client.extended_config.update(extended_config)
        return client

    def delete_user_pool_client(self, user_pool_id, client_id):
        user_pool = self.user_pools.get(user_pool_id)
        if not user_pool:
            raise ResourceNotFoundError(user_pool_id)

        if client_id not in user_pool.clients:
            raise ResourceNotFoundError(client_id)

        del user_pool.clients[client_id]

    # Identity provider
    def create_identity_provider(self, user_pool_id, name, extended_config):
        user_pool = self.user_pools.get(user_pool_id)
        if not user_pool:
            raise ResourceNotFoundError(user_pool_id)

        identity_provider = CognitoIdpIdentityProvider(name, extended_config)
        user_pool.identity_providers[name] = identity_provider
        return identity_provider

    @paginate(60)
    def list_identity_providers(self, user_pool_id, max_results=None, next_token=None):
        user_pool = self.user_pools.get(user_pool_id)
        if not user_pool:
            raise ResourceNotFoundError(user_pool_id)

        return user_pool.identity_providers.values()

    def describe_identity_provider(self, user_pool_id, name):
        user_pool = self.user_pools.get(user_pool_id)
        if not user_pool:
            raise ResourceNotFoundError(user_pool_id)

        identity_provider = user_pool.identity_providers.get(name)
        if not identity_provider:
            raise ResourceNotFoundError(name)

        return identity_provider

    def update_identity_provider(self, user_pool_id, name, extended_config):
        user_pool = self.user_pools.get(user_pool_id)
        if not user_pool:
            raise ResourceNotFoundError(user_pool_id)

        identity_provider = user_pool.identity_providers.get(name)
        if not identity_provider:
            raise ResourceNotFoundError(name)

        identity_provider.extended_config.update(extended_config)

        return identity_provider

    def delete_identity_provider(self, user_pool_id, name):
        user_pool = self.user_pools.get(user_pool_id)
        if not user_pool:
            raise ResourceNotFoundError(user_pool_id)

        if name not in user_pool.identity_providers:
            raise ResourceNotFoundError(name)

        del user_pool.identity_providers[name]

    # Group
    def create_group(self, user_pool_id, group_name, description, role_arn, precedence):
        user_pool = self.user_pools.get(user_pool_id)
        if not user_pool:
            raise ResourceNotFoundError(user_pool_id)

        group = CognitoIdpGroup(user_pool_id, group_name, description, role_arn, precedence)
        if group.group_name in user_pool.groups:
            raise GroupExistsException("A group with the name already exists")
        user_pool.groups[group.group_name] = group

        return group

    def get_group(self, user_pool_id, group_name):
        user_pool = self.user_pools.get(user_pool_id)
        if not user_pool:
            raise ResourceNotFoundError(user_pool_id)

        if group_name not in user_pool.groups:
            raise ResourceNotFoundError(group_name)

        return user_pool.groups[group_name]

    def list_groups(self, user_pool_id):
        user_pool = self.user_pools.get(user_pool_id)
        if not user_pool:
            raise ResourceNotFoundError(user_pool_id)

        return user_pool.groups.values()

    def delete_group(self, user_pool_id, group_name):
        user_pool = self.user_pools.get(user_pool_id)
        if not user_pool:
            raise ResourceNotFoundError(user_pool_id)

        if group_name not in user_pool.groups:
            raise ResourceNotFoundError(group_name)

        group = user_pool.groups[group_name]
        for user in group.users:
            user.groups.remove(group)

        del user_pool.groups[group_name]

    def admin_add_user_to_group(self, user_pool_id, group_name, username):
        group = self.get_group(user_pool_id, group_name)
        user = self.admin_get_user(user_pool_id, username)

        group.users.add(user)
        user.groups.add(group)

    def list_users_in_group(self, user_pool_id, group_name):
        group = self.get_group(user_pool_id, group_name)
        return list(group.users)

    def admin_list_groups_for_user(self, user_pool_id, username):
        user = self.admin_get_user(user_pool_id, username)
        return list(user.groups)

    def admin_remove_user_from_group(self, user_pool_id, group_name, username):
        group = self.get_group(user_pool_id, group_name)
        user = self.admin_get_user(user_pool_id, username)

        group.users.discard(user)
        user.groups.discard(group)

    # User
    def admin_create_user(self, user_pool_id, username, temporary_password, attributes):
        user_pool = self.user_pools.get(user_pool_id)
        if not user_pool:
            raise ResourceNotFoundError(user_pool_id)

        user = CognitoIdpUser(user_pool_id, username, temporary_password, UserStatus["FORCE_CHANGE_PASSWORD"], attributes)
        user_pool.users[user.username] = user
        return user

    def admin_get_user(self, user_pool_id, username):
        user_pool = self.user_pools.get(user_pool_id)
        if not user_pool:
            raise ResourceNotFoundError(user_pool_id)

        if username not in user_pool.users:
            raise UserNotFoundError(username)

        return user_pool.users[username]

    @paginate(60, "pagination_token", "limit")
    def list_users(self, user_pool_id, pagination_token=None, limit=None):
        user_pool = self.user_pools.get(user_pool_id)
        if not user_pool:
            raise ResourceNotFoundError(user_pool_id)

        return user_pool.users.values()

    def admin_disable_user(self, user_pool_id, username):
        user = self.admin_get_user(user_pool_id, username)
        user.enabled = False

    def admin_enable_user(self, user_pool_id, username):
        user = self.admin_get_user(user_pool_id, username)
        user.enabled = True

    def admin_delete_user(self, user_pool_id, username):
        user_pool = self.user_pools.get(user_pool_id)
        if not user_pool:
            raise ResourceNotFoundError(user_pool_id)

        if username not in user_pool.users:
            raise UserNotFoundError(username)

        user = user_pool.users[username]
        for group in user.groups:
            group.users.remove(user)

        del user_pool.users[username]

    def _log_user_in(self, user_pool, client, username):
        refresh_token = user_pool.create_refresh_token(client.id, username)
        access_token, id_token, expires_in = user_pool.create_tokens_from_refresh_token(refresh_token)

        return {
            "AuthenticationResult": {
                "IdToken": id_token,
                "AccessToken": access_token,
                "RefreshToken": refresh_token,
                "ExpiresIn": expires_in,
            }
        }

    def admin_initiate_auth(self, user_pool_id, client_id, auth_flow, auth_parameters):
        user_pool = self.user_pools.get(user_pool_id)
        if not user_pool:
            raise ResourceNotFoundError(user_pool_id)

        client = user_pool.clients.get(client_id)
        if not client:
            raise ResourceNotFoundError(client_id)

        if auth_flow == "ADMIN_NO_SRP_AUTH":
            username = auth_parameters.get("USERNAME")
            password = auth_parameters.get("PASSWORD")
            user = user_pool.users.get(username)
            if not user:
                raise UserNotFoundError(username)

            if user.password != password:
                raise NotAuthorizedError(username)

            if user.status == UserStatus["FORCE_CHANGE_PASSWORD"]:
                session = str(uuid.uuid4())
                self.sessions[session] = user_pool

                return {
                    "ChallengeName": "NEW_PASSWORD_REQUIRED",
                    "ChallengeParameters": {},
                    "Session": session,
                }

            return self._log_user_in(user_pool, client, username)
        elif auth_flow == "REFRESH_TOKEN":
            refresh_token = auth_parameters.get("REFRESH_TOKEN")
            id_token, access_token, expires_in = user_pool.create_tokens_from_refresh_token(refresh_token)

            return {
                "AuthenticationResult": {
                    "IdToken": id_token,
                    "AccessToken": access_token,
                    "ExpiresIn": expires_in,
                }
            }
        else:
            return {}

    def respond_to_auth_challenge(self, session, client_id, challenge_name, challenge_responses):
        user_pool = self.sessions.get(session)
        if not user_pool:
            raise ResourceNotFoundError(session)

        client = user_pool.clients.get(client_id)
        if not client:
            raise ResourceNotFoundError(client_id)

        if challenge_name == "NEW_PASSWORD_REQUIRED":
            username = challenge_responses.get("USERNAME")
            new_password = challenge_responses.get("NEW_PASSWORD")
            user = user_pool.users.get(username)
            if not user:
                raise UserNotFoundError(username)

            user.password = new_password
            user.status = UserStatus["CONFIRMED"]
            del self.sessions[session]

            return self._log_user_in(user_pool, client, username)
        else:
            return {}

    def confirm_forgot_password(self, client_id, username, password):
        for user_pool in self.user_pools.values():
            if client_id in user_pool.clients and username in user_pool.users:
                user_pool.users[username].password = password
                break
        else:
            raise ResourceNotFoundError(client_id)

    def change_password(self, access_token, previous_password, proposed_password):
        for user_pool in self.user_pools.values():
            if access_token in user_pool.access_tokens:
                _, username = user_pool.access_tokens[access_token]
                user = user_pool.users.get(username)
                if not user:
                    raise UserNotFoundError(username)

                if user.password != previous_password:
                    raise NotAuthorizedError(username)

                user.password = proposed_password
                if user.status == UserStatus["FORCE_CHANGE_PASSWORD"]:
                    user.status = UserStatus["CONFIRMED"]

                break
        else:
            raise NotAuthorizedError(access_token)

    def admin_update_user_attributes(self, user_pool_id, username, attributes):
        user_pool = self.user_pools.get(user_pool_id)
        if not user_pool:
            raise ResourceNotFoundError(user_pool_id)

        if username not in user_pool.users:
            raise UserNotFoundError(username)

        user = user_pool.users[username]
        user.update_attributes(attributes)


cognitoidp_backends = {}
for region in boto.cognito.identity.regions():
    cognitoidp_backends[region.name] = CognitoIdpBackend(region.name)


# Hack to help moto-server process requests on localhost, where the region isn't
# specified in the host header. Some endpoints (change password, confirm forgot
# password) have no authorization header from which to extract the region.
def find_region_by_value(key, value):
    for region in cognitoidp_backends:
        backend = cognitoidp_backends[region]
        for user_pool in backend.user_pools.values():
            if key == "client_id" and value in user_pool.clients:
                return region

            if key == "access_token" and value in user_pool.access_tokens:
                return region

    return cognitoidp_backends.keys()[0]
